import copy
import logging
from logging import FileHandler
import multiprocessing
import sys
from typing import Optional
import httpx

import http.client as httplib


class Configuration:
    """This class contains various settings of the API client.

        :param api_key: Dict to store API key(s).
          Each entry in the dict specifies an API key.
          The dict key is the name of the security scheme in the OAS specification.
          The dict value is the API key secret.
        :param api_key_prefix: Dict to store API prefix (e.g. Bearer).
          The dict key is the name of the security scheme in the OAS specification.
          The dict value is an API key prefix when generating the auth data.
        :param access_token: Access token.
        :param retries: Number of retries for API requests.

        :Example:

        API Key Authentication Example.
        Given the following security scheme in the OpenAPI specification:
          components:
            securitySchemes:
              cookieAuth:         # name for the security scheme
                type: apiKey
                in: cookie
                name: JSESSIONID  # cookie name

        You can programmatically set the cookie:

    conf = komodo.Configuration(
        api_key={'cookieAuth': 'abc123'}
        api_key_prefix={'cookieAuth': 'JSESSIONID'}
    )

        The following cookie will be added to the HTTP request:
           Cookie: JSESSIONID abc123
    """

    _default = None

    def __init__(
        self,
        api_key=None,
        api_key_prefix=None,
        access_token=None,
        retries=None,
    ) -> None:
        """Constructor"""
        self.temp_folder_path = None
        """Temp file folder for downloading files
        """
        # Authentication Settings
        self.api_key = {}
        if api_key:
            self.api_key = api_key
        """dict to store API key(s)
        """
        self.api_key_prefix = {}
        if api_key_prefix:
            self.api_key_prefix = api_key_prefix
        """dict to store API prefix (e.g. Bearer)
        """
        self.refresh_api_key_hook = None
        """function hook to refresh API key if expired
        """
        self.access_token = access_token
        """Access token
        """
        self.logger = {}
        """Logging Settings
        """
        self.logger["package_logger"] = logging.getLogger("komodo")
        self.logger["urllib3_logger"] = logging.getLogger("urllib3")

        self.verify_ssl = True
        """SSL/TLS verification
           Set this to false to skip verifying SSL certificate when calling API
           from https server.
        """
        self.safe_chars_for_path_param = ""
        """Safe chars for path_param
        """
        self.retries = retries
        """Adding retries to override default value 3
        """

    @classmethod
    def set_default(cls, default):
        """Set default instance of configuration.

        It stores default configuration, which can be
        returned by get_default_copy method.

        :param default: object of Configuration
        """
        cls._default = default

    @classmethod
    def get_default_copy(cls):
        """Deprecated. Please use `get_default` instead.

        Deprecated. Please use `get_default` instead.

        :return: The configuration object.
        """
        return cls.get_default()

    @classmethod
    def get_default(cls):
        """Return the default configuration.

        This method returns newly created, based on default constructor,
        object of Configuration class or returns a copy of default
        configuration.

        :return: The configuration object.
        """
        if cls._default is None:
            cls._default = Configuration()
        return cls._default

    def get_api_key_with_prefix(self, identifier, alias=None):
        """Gets API key (with prefix if set).

        :param identifier: The identifier of apiKey.
        :param alias: The alternative identifier of apiKey.
        :return: The token for api key authentication.
        """
        if self.refresh_api_key_hook is not None:
            self.refresh_api_key_hook(self)
        key = self.api_key.get(identifier, self.api_key.get(alias) if alias is not None else None)
        if key:
            prefix = self.api_key_prefix.get(identifier)
            if prefix:
                return "%s %s" % (prefix, key)
            else:
                return key

    def auth_settings(self):
        """Gets Auth Settings dict for api client.

        :return: The Auth Settings information dict.
        """
        auth = {}
        if self.access_token is not None:
            auth["HTTPBearer"] = {"type": "bearer", "in": "header", "key": "Authorization", "value": self.get_api_key_with_prefix("HTTPBearer")}
        if "x-account-id" in self.api_key:
            auth["x-account-id"] = {
                "type": "api_key",
                "in": "header",
                "key": "x-account-id",
                "value": self.get_api_key_with_prefix(
                    "x-account-id",
                ),
            }
        return auth



from typing import Any

import httpx
import pyperclip
import typer
from questionary import Choice, Style, select
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.text import Text

from cli.util import ENVIRONMENT_OPTION, get_headers,print_table, resolve_environment


def fetch_accounts(token: str, base_url: str) -> dict[str, str]:
    url = f"{base_url}/v1/iam/users/me?expand=organizations.accounts"
    headers = get_headers(token)
    with httpx.Client() as client:
        resp = client.get(url, headers=headers)
        data = resp.json()

    accounts = {}
    for org in data.get("organizations", []):
        for acct in org.get("accounts", []):
            accounts[acct["slug"]] = acct["account_id"]
    return accounts


def fetch_accounts_detailed(token: str, base_url: str) -> list[dict[str, Any]]:
    """Fetch all accounts with complete metadata."""
    url = f"{base_url}/v1/iam/users/me?expand=organizations.accounts"
    headers = get_headers(token)

    with httpx.Client() as client:
        resp = client.get(url, headers=headers)
        data = resp.json()

    accounts = []
    for org in data.get("organizations", []):
        org_name = org.get("name", "")
        for acct in org.get("accounts", []):
            # Add organization name to account metadata
            acct_data = {**acct, "organization_name": org_name}
            accounts.append(acct_data)
    return accounts


account_app = typer.Typer(short_help="Get and set the Komodo Account.", no_args_is_help=True)


@account_app.command("set")
@resolve_environment
def account_set(environment: str = ENVIRONMENT_OPTION):
    """Set the Komodo Account."""
    from komodo.config import SDKSettings
    from komodo.auth.Session import Session, CredentialsFileDefaultProvider
    env = environment.lower()

    settings = SDKSettings(environment=env)
    console = Console()
    session = Session(environment=env)
    credentials = session.load_credentials()
    if not credentials.access_token:
        console.print(Text("No access token found. Please login first.", style="bold red"))
        raise typer.Exit(code=1)

    with Progress(SpinnerColumn(), TextColumn("[progress.description]{task.description}"), transient=True) as progress:
        task = progress.add_task("Fetching accounts...", start=True)
        try:
            accounts = fetch_accounts(session.access_token, settings.komodo_external)
        except Exception as e:
            progress.stop()

            if "nodename nor servname provided, or not known" in str(e):
                console.print(Text("You are not connected to Twingate. Please connect to Twingate and try again.", style="bold red"))
            else:
                console.print(Text(f"Failed to fetch accounts: {e}", style="bold red"))

            raise typer.Exit(code=1)
        progress.update(task, description="Accounts fetched!")

    if not accounts:
        console.print(Text(f"No accounts found for this user in {env}.", style="bold red"))
        raise typer.Exit(code=1)
    max_slug_len = max(len(slug) for slug in accounts)
    choices = []
    for slug, account_id in accounts.items():
        label = f"{slug.ljust(max_slug_len)}  |  {account_id}"
        value = f"{slug}|{account_id}"
        choices.append(Choice(title=label, value=value))
    custom_style = Style(
        [
            ("qmark", "fg:#00b7c2 bold"),
            ("question", "bold"),
            ("answer", "fg:#f44336 bold"),
            ("pointer", "fg:#00b7c2 bold"),
            ("highlighted", "fg:#00b7c2 bold"),
            ("selected", "fg:#00b7c2 bold"),
            ("separator", "fg:#cc5454"),
            ("instruction", ""),
            ("text", ""),
            ("disabled", "fg:#858585 italic"),
        ]
    )
    selected = select("Select an account:", choices=choices, use_shortcuts=False, style=custom_style).ask()
    if not selected:
        console.print(Text("No account selected. Exiting.", style="bold red"))
        raise typer.Exit(code=1)
    selected_slug, selected_account_id = selected.split("|", 1)
    selected_slug = selected_slug.strip()
    selected_account_id = selected_account_id.strip()
    selected_account_id = str(selected_account_id)

    provider = CredentialsFileDefaultProvider(settings)
    file_path = provider.set_credentials(env, {"account_id": selected_account_id, "account_slug": selected_slug})

    console.print(Text(f"✏️ Account set to: {selected_slug} ({selected_account_id}) in {file_path}", style="bold green"))


@account_app.command("get")
@resolve_environment
def account_get(environment: str = ENVIRONMENT_OPTION):
    """Copy the set komodo account id to clipboard."""
    from komodo.config import SDKSettings
    from komodo.auth.Session import CredentialsFileDefaultProvider
    console = Console()
    env = environment.lower()
    settings = SDKSettings(environment=env)
    provider = CredentialsFileDefaultProvider(settings)
    creds = provider.get_credentials()
    account_id = getattr(creds, "account_id", None)
    account_slug = getattr(creds, "account_slug", None)
    if not account_id or not account_slug:
        console.print(Text("KOMODO_ACCOUNT_ID or KOMODO_ACCOUNT_SLUG not set. Please run 'komodo account set' first.", style="bold red"))
        raise typer.Exit(code=1)
    try:
        pyperclip.copy(str(account_id))
        console.print(Text(f"📎 Copied account id for {account_slug} ({account_id}) into clipboard ", style="bold green"))
    except Exception:
        console.print(Text("📎Could not copy to clipboard. Please copy manually.", style="bold red"))
        console.print(Text(f"Account id for {account_slug}: {account_id}", style="bold yellow"))


@account_app.command("list")
@resolve_environment
def account_list(environment: str = ENVIRONMENT_OPTION):
    """List the komodo accounts"""
    from komodo.config import SDKSettings
    import komodo.auth.Session as session
    env = environment.lower()

    settings = SDKSettings(environment=env)
    console = Console()
    session = session.Session(environment=env)
    credentials = session.load_credentials()
    if not credentials.access_token:
        console.print(Text("No access token found. Please login first.", style="bold red"))
        raise typer.Exit(code=1)

    with Progress(SpinnerColumn(), TextColumn("[progress.description]{task.description}"), transient=True) as progress:
        task = progress.add_task("Fetching accounts...", start=True)
        try:
            accounts = fetch_accounts_detailed(session.access_token, settings.komodo_external)
        except Exception as e:
            progress.stop()

            if "nodename nor servname provided, or not known" in str(e):
                console.print(Text("You are not connected to Twingate. Please connect to Twingate and try again.", style="bold red"))
            else:
                console.print(Text(f"Failed to fetch accounts: {e}", style="bold red"))

            raise typer.Exit(code=1)
        progress.update(task, description="Accounts fetched!")

    print_table(accounts, "Komodo Accounts", priority_keys=["slug", "account_id", "name", "organization_name"], console=console)




import os
from functools import wraps

import httpx
import typer

from typing import Optional, List, Dict, Any

from rich.console import Console
from rich.table import Table
from rich.text import Text


ENVIRONMENT_OPTION = typer.Option(
    None,
    "--environment",
    "-E",
    help="Specify the environment (e.g., production or integration). Defaults to production.",
)

def resolve_environment(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if "environment" in kwargs:
            kwargs["environment"] = get_environment(kwargs["environment"])
        return func(*args, **kwargs)

    return wrapper


def get_environment(environment_flag: str, default: str = "production") -> str:
    """
    Get the environment setting.
    Priority: environment_flag > KOMODO_ENVIRONMENT env var > 'production' default
    """

    return environment_flag or os.getenv("KOMODO_ENVIRONMENT") or default


def get_session_and_settings(environment: str) -> tuple:
    """Get authenticated session, settings, and console for API calls"""
    from komodo.config import SDKSettings
    from komodo.auth.Session import Session
    env = environment.lower()
    settings = SDKSettings(environment=env)
    console = Console()
    session = Session(environment=env)

    credentials = session.load_credentials()
    if not credentials:
        console.print(Text("No access token found. Please login first.", style="bold red"))
        raise typer.Exit(code=1)

    return session, settings, console, credentials


def handle_api_error(response: httpx.Response, console: Console, operation: str, custom_error_messages: Optional[Dict[int, str]] = None) -> None:
    """Handle API errors with proper error messages"""
    if custom_error_messages and response.status_code in custom_error_messages:
        console.print(Text(custom_error_messages[response.status_code], style="bold red"))
        raise typer.Exit(code=1)

    if response.status_code == 401:
        console.print(Text("Authentication failed. Please run 'komodo login' first.", style="bold red"))
        raise typer.Exit(code=1)
    elif response.status_code == 403:
        console.print(Text("Access denied. You don't have permission to perform this operation.", style="bold red"))
        raise typer.Exit(code=1)
    elif response.status_code == 404:
        console.print(Text("Resource not found.", style="bold red"))
        raise typer.Exit(code=1)
    elif response.status_code == 422:
        try:
            error_data = response.json()
            console.print(Text(f"Validation error: {error_data}", style="bold red"))
        except:
            console.print(Text("Validation error occurred.", style="bold red"))
        raise typer.Exit(code=1)
    else:
        console.print(Text(f"API error during {operation}: {response.status_code} - {response.text}", style="bold red"))
        raise typer.Exit(code=1)


def get_log_content(url: str, console: Console) -> str:
    """Fetch a presigned URL from a given URL"""
    with httpx.Client() as client:
        try:
            return client.get(url).text
        except httpx.RequestError as e:
            console.print(Text(f"Network error during fetch_presigned_url: {e}", style="bold red"))
            raise typer.Exit(code=1)


def make_api_request(method: str, url: str, headers: Dict[str, str], json_data: Optional[Dict[str, Any]] = None, console: Console = None, operation: str = "API call", custom_error_messages: Optional[Dict[int, str]] = None) -> Dict[str, Any]:
    """Make an API request with proper error handling"""
    if console is None:
        console = Console()
    with httpx.Client() as client:
        try:
            if method.upper() == "GET":
                response = client.get(url, headers=headers)
            elif method.upper() == "POST":
                response = client.post(url, headers=headers, json=json_data)
            elif method.upper() == "PUT":
                response = client.put(url, headers=headers, json=json_data)
            elif method.upper() == "DELETE":
                response = client.delete(url, headers=headers)
            else:
                raise ValueError(f"Unsupported HTTP method: {method}")

            if response.status_code == 200:
                return response.json()
            else:
                handle_api_error(response, console, operation, custom_error_messages)

        except httpx.RequestError as e:
            console.print(Text(f"Network error during {operation}: {e}, {url}", style="bold red"))
            raise typer.Exit(code=1)


def get_apps_api_url(environment: str) -> str:
    """
    Get the API URL for the given environment.

    Priority: KOMODO_ENVIRONMENT env var > command flag > 'production' default
    """

    if environment == "integration":
        return "https://developer-platform-apps.internal.staging.onkomodo.com/apps-integration/v1/"
    else:
        return "https://api.khinternal.net/apps-production/v1/"


def get_platform_infrastructure_api_url(environment: str) -> str:
    """Get the API URL for the given environment"""
    if environment == "integration":
        return "https://developer-platform-infra.internal.dev.onkomodo.com/v1"
    else:
        return "https://developer-platform-infra.internal.onkomodo.com/v1"


def print_table(data: List[Dict[str, Any]], title: str, priority_keys: List[str] = [], console: Console = None) -> None:
    if console is None:
        console = Console()
    if not data:
        console.print(Text(f"No {title} found.", style="bold yellow"))
        return

    table = Table(title=title, show_header=True, header_style="bold cyan")

    # Dynamically add columns based on the keys in the first account
    # Sort keys to have a consistent column order
    all_keys = set()
    for item in data:
        all_keys.update(item.keys())

    # Order columns: priority keys first, then remaining keys alphabetically
    ordered_keys = [k for k in priority_keys if k in all_keys]
    remaining_keys = sorted([k for k in all_keys if k not in priority_keys])
    column_keys = ordered_keys + remaining_keys

    # Add columns to table
    for key in column_keys:
        # Format column names nicely
        column_name = key.replace("_", " ").title()
        table.add_column(column_name, style="green")

    # Add rows
    for item in data:
        row_values = [str(item.get(key, "")) for key in column_keys]
        table.add_row(*row_values)

    console.print(table)

def get_headers(access_token: str, account_id: str | None = None) -> dict[str, str]:
    headers = {
        "Authorization": f"Bearer {access_token}"
    }
    if account_id:
        headers["x-account-id"] = account_id
    if "KOMODO_SESSION_ID" in os.environ:
        headers["x-session-id"] = os.environ["KOMODO_SESSION_ID"]
    if "KOMODO_USER_ID" in os.environ:
        headers["x-user-id"] = os.environ["KOMODO_USER_ID"]
    if "KOMODO_USER_TYPE" in os.environ:
        headers["x-user-type"] = os.environ["KOMODO_USER_TYPE"]
    if "KOMODO_ORGANIZATION_ID" in os.environ:
        headers["x-organization-id"] = os.environ["KOMODO_ORGANIZATION_ID"]
    if "KOMODO_DEVICE_ID" in os.environ:
        headers["x-device-id"] = os.environ["KOMODO_DEVICE_ID"]
    return headers



from typing import Any, Optional

from komodo._vendored.komodo_connector.connection import Connection
from komodo._vendored.komodo_connector.constants import (
    DEV_KOMODO_PROXY_SERVICE_DEFAULT_PORT,
    SNOWFLAKE_MAIN_ACCOUNTS,
    VENDOR_CONNECTION_STRING_MAP,
    Vendors,
)
from komodo._vendored.komodo_connector.models import (
    KomodoAuthenticationParams,
)
from komodo._vendored.komodo_connector.setup_driver import (
    load_driver,
    parse_komodo_url,
)


def connect(
    connection_string: Optional[str] = None,
    *,
    vendor: Vendors = Vendors.SNOWFLAKE,
    account_id: Optional[str] = None,
    profile_id: Optional[str] = None,
    secret: Optional[str] = None,
    port: Optional[str] = None,
    protocol: Optional[str] = None,
    environment: Optional[str] = None,
    komodo_host: Optional[str] = None,
    app_id: Optional[str] = None,
    **kwargs: Optional[Any],
) -> Connection:
    """
    Connects to a Komodo database using the provided connection parameters.

    Args:
        connection_string (Optional[str]): The connection string. Defaults to None.
        vendor (Vendors): The vendor of the Komodo database. Defaults to Vendors.SNOWFLAKE.
        account_id (Optional[str]): The account ID. Defaults to None.
        profile_id (Optional[str]): The profile ID. Defaults to None.
        secret (Optional[str]): The secret. Defaults to None.
        port (Optional[str]): The port number. Defaults to None.
        protocol (Optional[str]): The protocol. Defaults to None.
        environment (Optional[str]): The environment. Defaults to None.
        komodo_host (Optional[str]): The Komodo host. Defaults to None.
        app_id (Optional[str]): The application ID. Defaults to None.
        **kwargs (Optional[Any]): Additional keyword arguments.

    Returns:
        Connection: The connection object.

    Raises:
        AssertionError: Raised when required parameters are not provided.

    Example:
        ```python
        connection_string = "komodo://..."
        connection = connect(connection_string)
        connection.execute("SELECT * FROM table")
        ```
    """

    if connection_string is None:
        assert account_id is not None, "account_id cannot be None"
        assert profile_id is not None, "profile_id cannot be None"
        assert secret is not None, "secret cannot be None"
        connection_string = create_connection_string(
            vendor=vendor,
            account_id=account_id,
            profile_id=profile_id,
            secret=secret,
            port=port,
            protocol=protocol,
            environment=environment,
            komodo_host=komodo_host,
            app_id=app_id,
        )

    vendor = connection_string.split("://", 1)[1].split("+", 1)[0]
    if vendor not in Vendors.values():
        raise ValueError(f"Invalid vendor: {vendor}")

    komodo_account_id = connection_string.split("@")[1].split(":")[0]

    # Temporary override to native snowflake connector for legacy connections
    if vendor == Vendors.SNOWFLAKE and komodo_account_id in SNOWFLAKE_MAIN_ACCOUNTS:
        connection_string = connection_string.replace(vendor, Vendors.SNOWFLAKE_NATIVE.value, 1)
    driver_name, connection_config = parse_komodo_url(connection_string)
    connection_config_dict = dict(connection_config)
    connection_config_dict["query_params"] = {**connection_config_dict.get("query_params", {}), **kwargs}
    driver, connection_creator = load_driver(driver_name)
    return connection_creator(driver, connection_config_dict).create()


def create_connection_string(
    *,
    vendor: Vendors,
    account_id: str,
    profile_id: str,
    secret: str,
    port: Optional[str],
    protocol: Optional[str],
    environment: Optional[str],
    komodo_host: Optional[str],
    app_id: Optional[str],
) -> str:
    """
    Creates a connection string for a given vendor, account, profile, secret, and optional connection parameters.

    Returns:
        str: The generated connection string.

    """

    auth_params = KomodoAuthenticationParams(
        komodo_account=account_id,
        komodo_user=profile_id,
        komodo_token=secret,
    )
    port = port or DEV_KOMODO_PROXY_SERVICE_DEFAULT_PORT
    return build_connection_string(
        vendor,
        auth_params,
        port=port,
        protocol=protocol,
        environment=environment,
        komodo_host=komodo_host,
        app_id=app_id,
    )


def build_connection_string(
    vendor: Vendors,
    auth_params: KomodoAuthenticationParams,
    port: Optional[str] = None,
    protocol: Optional[str] = None,
    environment: Optional[str] = None,
    komodo_host: Optional[str] = None,
    app_id: Optional[str] = None,
) -> str:
    """
    Creates a connection string for a Komodo database based on the provided parameters.

    Args:
        vendor (Vendors): The vendor of the Komodo database.
        auth_params (KomodoAuthenticationParams): The authentication parameters.
        port (Optional[str]): The port number. Defaults to None.
        protocol (Optional[str]): The protocol. Defaults to None.
        environment (Optional[str]): The environment. Defaults to None.
        komodo_host (Optional[str]): The Komodo host. Defaults to None.

    Returns:
        str: The generated connection string.
    """

    connection_string = f"komodo://{VENDOR_CONNECTION_STRING_MAP[vendor]}+{auth_params.komodo_user}:{auth_params.komodo_token}@{auth_params.komodo_account}:{port}"

    query_params = []
    if protocol is not None:
        query_params.append(f"protocol={protocol}")
    if environment is not None:
        query_params.append(f"env={environment}")
    if komodo_host is not None:
        query_params.append(f"komodo_host={komodo_host}")
    if app_id is not None:
        query_params.append(f"app_id={app_id}")

    if query_params:
        connection_string += "?" + "&".join(query_params)
    return connection_string



from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from komodo.iam.models.core.membership_type import MembershipType
from komodo.iam.models.core.user_lookup_organization import UserLookupOrganization
from komodo.iam.models.core.user_status import UserStatus
from komodo.iam.models.core.user_type import UserType
from typing import Optional, Set, Union
from typing_extensions import Self


class UserLookupResponse(BaseModel):
    """
    Details of User that may contain expanded information about Organizations, Groups, and Accounts the User is a member of.
    """  # noqa: E501

    auth_tenant_id: Optional[StrictStr] = Field(default=None, description="the id of the user that is issued by the auth tenant")
    created_at: datetime = Field(description="Time of creation")
    created_by: Optional[StrictStr] = Field(default=None, description="Created By")
    display_name: Optional[StrictStr] = Field(default=None, description="Display name (SCIM)")
    email: StrictStr = Field(description="User email")
    family_name: Optional[StrictStr] = Field(default=None, description="Family name (SCIM)")
    formatted: Optional[StrictStr] = Field(default=None, description="Formatted name (SCIM)")
    given_name: Optional[StrictStr] = Field(default=None, description="Given name (SCIM)")
    honorific_prefix: Optional[StrictStr] = Field(default=None, description="Honorific prefix (SCIM)")
    honorific_suffix: Optional[StrictStr] = Field(default=None, description="Honorific suffix (SCIM)")
    last_account_used: Optional[StrictStr] = Field(default=None, description="Account id")
    locale: Optional[StrictStr] = Field(default=None, description="Locale (SCIM)")
    membership_type: Optional[Union[MembershipType, str]] = Field(default=None, description="The user's level of access across the entire platform-rbac system")
    middle_name: Optional[StrictStr] = Field(default=None, description="Middle name (SCIM)")
    nick_name: Optional[StrictStr] = Field(default=None, description="Nick name (SCIM)")
    organizations: Optional[List[UserLookupOrganization]] = None
    preferred_language: Optional[StrictStr] = Field(default=None, description="Preferred language (SCIM)")
    profile_url: Optional[StrictStr] = Field(default=None, description="Profile URL (SCIM)")
    status: Union[UserStatus, str] = Field(description="Status (active or inactive)")
    timezone: Optional[StrictStr] = Field(default=None, description="Timezone (SCIM)")
    title: Optional[StrictStr] = Field(default=None, description="Title (SCIM)")
    updated_at: datetime = Field(description="Time of last update")
    updated_by: Optional[StrictStr] = Field(default=None, description="Updated By")
    user_id: StrictStr = Field(description="User ID")
    user_type: Union[UserType, str] = Field(description="User type")
    __properties: ClassVar[List[str]] = [
        "auth_tenant_id",
        "created_at",
        "created_by",
        "display_name",
        "email",
        "family_name",
        "formatted",
        "given_name",
        "honorific_prefix",
        "honorific_suffix",
        "last_account_used",
        "locale",
        "membership_type",
        "middle_name",
        "nick_name",
        "organizations",
        "preferred_language",
        "profile_url",
        "status",
        "timezone",
        "title",
        "updated_at",
        "updated_by",
        "user_id",
        "user_type",
    ]

    model_config = ConfigDict(populate_by_name=True, validate_assignment=True, protected_namespaces=(), extra="allow")

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of UserLookupResponse from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in organizations (list)
        _items = []
        if self.organizations:
            for _item_organizations in self.organizations:
                if _item_organizations:
                    _items.append(_item_organizations.to_dict())
            _dict["organizations"] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of UserLookupResponse from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "auth_tenant_id": obj.get("auth_tenant_id"),
                "created_at": obj.get("created_at"),
                "created_by": obj.get("created_by"),
                "display_name": obj.get("display_name"),
                "email": obj.get("email"),
                "family_name": obj.get("family_name"),
                "formatted": obj.get("formatted"),
                "given_name": obj.get("given_name"),
                "honorific_prefix": obj.get("honorific_prefix"),
                "honorific_suffix": obj.get("honorific_suffix"),
                "last_account_used": obj.get("last_account_used"),
                "locale": obj.get("locale"),
                "membership_type": obj.get("membership_type"),
                "middle_name": obj.get("middle_name"),
                "nick_name": obj.get("nick_name"),
                "organizations": [UserLookupOrganization.from_dict(_item) for _item in obj["organizations"]] if obj.get("organizations") is not None else None,
                "preferred_language": obj.get("preferred_language"),
                "profile_url": obj.get("profile_url"),
                "status": obj.get("status"),
                "timezone": obj.get("timezone"),
                "title": obj.get("title"),
                "updated_at": obj.get("updated_at"),
                "updated_by": obj.get("updated_by"),
                "user_id": obj.get("user_id"),
                "user_type": obj.get("user_type"),
            }
        )
        return _obj



# import apis into sdk package
from komodo.api.iam_api import IamApi

# import ApiClient
from komodo.api_response import ApiResponse
from komodo.api_client import ApiClient
from komodo.configuration import Configuration
from komodo.exceptions import OpenApiException
from komodo.exceptions import ApiTypeError
from komodo.exceptions import ApiValueError
from komodo.exceptions import ApiKeyError
from komodo.exceptions import ApiAttributeError
from komodo.exceptions import ApiException

# import models into sdk package
from komodo.core.models.error_response import ErrorResponse
from komodo.core.models.http_validation_error import HTTPValidationError


from komodo.iam.models.core.membership_type import MembershipType
from komodo.iam.models.core.private_role import PrivateRole
from komodo.iam.models.core.role_response import RoleResponse
from komodo.iam.models.core.service_principal_response import ServicePrincipalResponse
from komodo.iam.models.core.user_lookup_account import UserLookupAccount
from komodo.iam.models.core.user_lookup_group import UserLookupGroup
from komodo.iam.models.core.user_lookup_organization import UserLookupOrganization
from komodo.iam.models.core.user_lookup_response import UserLookupResponse
from komodo.iam.models.core.user_status import UserStatus
from komodo.iam.models.core.user_type import UserType

from komodo.iam.models.identity.identity_response import IdentityResponse
from komodo.iam.models.identity.jwt_type import JWTType
from komodo.iam.models.service_principals.service_principal_token_create import ServicePrincipalTokenCreate
from komodo.iam.models.service_principals.service_principal_token_response import ServicePrincipalTokenResponse
from komodo.iam.models.service_principals.service_principal_update import ServicePrincipalUpdate

# import Client
from komodo.client import Client

from komodo.snowflake import *


from enum import Enum
from typing import Optional
from uuid import UUID

from pydantic import BaseModel, Field, RootModel,field_validator
from typing import Union, Annotated, Literal, List
from packaging import version


class EnvEnum(str, Enum):
    local = "local"
    dev = "dev"
    test = "test"
    qa = "qa"
    stage = "stage"
    integration = "integration"
    prod = "prod"

class Trace(BaseModel):
    """A class used to represent a Trace (cf definition here https://opentelemetry.io/docs/concepts/signals/traces/)

    Attributes
    ----------
    trace_id: str
        Represents the trace id
    parent_id: str
        Represents the parent id
    """
    
    trace_id: str
    parent_id: str

class BasicProperties(BaseModel):
    """A class used to represent a service or an application

    Attributes
    ----------
    name: str
        Represents the name of the service
    properties: dict
        Represents any additional properties. Could be empty but not None

    """

    name: str
    properties: dict

class VersionedProperties(BasicProperties):
    """A class used to represent a service or an application

    Attributes
    ----------
    name: str
        Represents the name of the object
    properties: dict
        Represents any additional properties. Could be empty but not None
    version: str
        Represents the version of the object. Must be a valid version
    """

    version: str
    
    @field_validator('version')
    def must_be_a_version(cls, v):
        try:
            version.parse(v)
        except:
            raise ValueError('must be a version, for more information: https://packaging.python.org/en/latest/specifications/version-specifiers/')
        return v

class RootSnowflakeTag(BaseModel, extra='forbid'):
    """A class used to represent a service or an application

    Attributes
    ----------
    version: str = "3.0.0"
        The version of the current schema. Please do not modify this value.
    type: str
        commercial or eng
    env: EnvEnum
        The environment of the executed query
    nest: str
        The nest that is requesting the execution of the query
    account_id: Optional[UUID]
        The account id coming from common RBAC service, Otherwise, the tenantId.
   """

    version: str = "3.0.0"
    env: EnvEnum
    nest: str
    account_id: Optional[UUID] = None    

class EngSnowflakeTag(RootSnowflakeTag):
    """A class used to represent a service or an application

    Attributes
    ----------
    user_id: Optional[UUID]
        The user id coming from common RBAC service.
    trace: Optional[Trace]
        The trace id and parent id of the query execution (could be None)
    service: BasicProperties
        The service that is issuing the query. Should match the APM name in New Relic 
        Relates to SERVICE_NAME attribute if using OTEL https://opentelemetry.io/docs/concepts/sdk-configuration/general-sdk-configuration/#otel_service_name
        or [app_name conf if using NR agent]https://docs.newrelic.com/docs/apm/agents/python-agent/configuration/python-agent-configuration/#app_name 
    client: BasicProperties
        The client application that the service is executing on its behalf (could be None) 
        It related to the "app-id". Also a version is needed 
    """
    type: Literal['eng']
    user_id: Optional[UUID] = None
    trace: Optional[Trace] = None
    service: VersionedProperties 
    client: VersionedProperties

class CommercialSnowflakeTag(RootSnowflakeTag):
    """A class used to represent a service or an application

    Attributes
    ----------

    subnest: str
        The subnest that is requesting the execution of the query
    opportunity_id: str
        The opportunity id from salesforce related to the query
    application: BasicProperties
        The application name for which the query is running 
    data: VersionedProperties
        The fabric that the application is executing the query 
    data_products: List[VersionedProperties]
        Multiple fabrics can be utilized across various applications executing queries
    time_tracking_id: str
        NetSuite OpenAir ID associated with services provided to a prospect/customer
    """
    type : Literal['commercial']
    subnest: Optional[str] = None
    opportunity_id: str
    application: BasicProperties
    data: VersionedProperties
    data_products: Optional[List[VersionedProperties]]
    time_tracking_id: str

class DataFoundationSnowflakeTag(RootSnowflakeTag):
    """A class used to track the production of data products by the Data Foundation liar

    Attributes
    ----------

    data_product: VersionedProperties
        The fabric that the application is executing the query
        examples:
            - {"name": "polyester", "version": "23.0.0", "data_type": "rx", "properties": {}}
            - {"name": "patient_demographics", "version": "23.0.0", "properties": {}}
            - {"name": "arbok", "version": "1.0.0", "stage": "raw", "properties": {}}
            - {"name": "arbok", "version": "23.0.0", "stage": "johto", "data_type": "rx", "properties": {}}
    """
    type : Literal["data_foundation"]
    nest: Literal["foundation", "context", "cdo", "data_ingestion", "events", "product", "quality"]
    data_product: VersionedProperties

class SnowflakeTag(RootModel):
    root: Annotated[Union[CommercialSnowflakeTag, EngSnowflakeTag, DataFoundationSnowflakeTag], Field(..., discriminator='type')]
    
    def dict(self, **kwargs):
        return super().model_dump(exclude_none=True, **kwargs)



import pyperclip
from datetime import datetime, timezone
import typer
from rich.console import Console
from rich.text import Text

from cli.account import fetch_accounts
from cli.util import resolve_environment, ENVIRONMENT_OPTION


@resolve_environment
def login(environment: str = ENVIRONMENT_OPTION):
    """Authenticate and save credentials."""
    from komodo.auth.Session import CredentialsFileDefaultProvider, Session
    from komodo.auth.constants import KOMODO_IDP_API_SCOPE
    import komodo.auth.OAuthFlows as oauthflows
    from komodo.config import SDKSettings
    console = Console()

    settings = SDKSettings(environment=environment)
    oauth = oauthflows.OAuthFlows(domain=settings.komodo_idp_domain, rbac_url=settings.komodo_external)
    try:
        tokens = oauth.device_authorization(settings.komodo_client_id, settings.komodo_idp_audience, KOMODO_IDP_API_SCOPE, rich_console=console)

        provider = CredentialsFileDefaultProvider(settings)
        file_path = provider.set_credentials(environment, tokens)

        session = Session(environment=environment)
        credentials = session.load_credentials()
        formatted_expires_at = datetime.fromtimestamp(int(tokens['expires_at']), tz=timezone.utc).strftime('%H:%M UTC on %d %B %Y')
        message = f"✏️ Success! Credentials written to: {file_path}\n Credentials valid until {formatted_expires_at}"
        if credentials.access_token:
            accounts = fetch_accounts(credentials.access_token, settings.komodo_external)
            if len(accounts) == 1:
                account_id = list(accounts.values())[0]
                account_slug = list(accounts.keys())[0]
                provider = CredentialsFileDefaultProvider(settings)
                provider.set_credentials(environment, {"account_id": account_id, "account_slug": account_slug})
                message += f"\n Using account: {account_slug} ({account_id})"
        console.print(Text(message, style="bold green"))
    except Exception as e:
        console.print(Text(f"Device authorization failed: {e}", style="bold red"))


@resolve_environment
def jwt(environment: str = ENVIRONMENT_OPTION):
    """Display access token and copy to clipboard."""
    from komodo.auth.Session import Session
    session = Session(environment=environment)
    console = Console()
    token = session.access_token
    if not token:
        console.print(Text("No access token found. Please `komodo login` first.", style="bold red"))
        raise typer.Exit(code=1)
    console.print(Text(token))
    try:
        pyperclip.copy(token)
        console.print(Text("📎 JWT copied to clipboard", style="bold green"))
    except Exception:
        console.print(Text("Could not copy to clipboard. Please copy manually.", style="bold red"))



import asyncio
import logging
import warnings

from komodo.client import Client
from komodo._vendored.komodo_connector.connector import connect
from komodo._vendored.komodo_connector.connection_creators.snowflake.cursor import KomodoSnowflakeCursor
from komodo._vendored.komodo_connector.connection import Connection
from snowflake.connector.cursor import SnowflakeCursor
from komodo.auth.Session import Session
from komodo._vendored.snowflake_query_tag.models import SnowflakeTag, VersionedProperties

logger = logging.getLogger(__name__)

# Suppress UserWarning from pandas when using komodo_connector's Connection object, since SQLAlchemy is
# technically the only DBAPI2.0 compliant connection supported by pandas.
warnings.simplefilter(action="ignore", category=UserWarning)


class UnsetAccountError(Exception):
    """Exception raised for errors in the snowflake connection"""


def get_snowflake_connection(account_id: str | None = None, jwt: str | None = None, client_id: str | None = None, client_secret: str | None = None, profile: str | None = None) -> Connection:
    if jwt is not None and (client_id is not None or client_secret is not None):
        raise ValueError("Cannot pass in both a JWT and client_id/client_secret.")
    
    if profile is not None and (jwt is not None or client_id is not None or client_secret is not None):
        raise ValueError("Cannot specify profile along with explicit credentials (jwt, client_id, or client_secret).")

    client_kwargs = {}

    if account_id is not None:
        client_kwargs["account_id"] = account_id

    if profile is not None:
        client_kwargs["auth_session"] = Session(profile=profile)
        if account_id is not None and client_kwargs["auth_session"].account_id is not None:
            raise ValueError(
             "The account_id parameter was provided, but the selected profile also contains an account_id. "
             "Please specify the account_id in only one location."
             )


    elif jwt is not None:
        client_kwargs["auth_session"] = Session(access_token=jwt)
    elif client_id is not None and client_secret is not None:
        client_kwargs["auth_session"] = Session(client_id=client_id, client_secret=client_secret)
    elif client_id is not None or client_secret is not None:
        raise ValueError("Invalid credentials provided. Please provide either a JWT or both client_id and client_secret.")

    client = Client(**client_kwargs)

    jwt = client.auth.access_token
    account_id = client.x_account_id

    if account_id is None:
        raise UnsetAccountError("An account_id was not passed and could not be read from the SDK client session. Please pass an account_id or set it in your credentials file.")

    me = client.iam.get_current_identity()
    identity_id = me.user.user_id if me.user else me.service_principal.service_principal_id

    environment = client.auth._settings.environment
    env = "prod" if environment == "production" else "dev"
    # TODO: We are in the process of getting final permission to remove the snowflake query tags
    app_id = "compute"
    client_type = "eng"
    nest = "mapkit"
    service = VersionedProperties(name="mapkit", version="1.0.0", properties={})
    client = VersionedProperties(name="python_sdk", version="1.0.0", properties={})
    snowflake_tag = SnowflakeTag(type=client_type, env=env, nest=nest, account_id=account_id, user_id=identity_id, service=service, client=client)
    query_tag = snowflake_tag.model_dump_json(exclude_none=True, exclude_unset=True, by_alias=True, indent=4)

    connection = connect(
        account_id=account_id,
        profile_id=identity_id,
        secret=f"Bearer {jwt}",
        app_id=app_id,
        protocol=None,
        environment=env,
        session_parameters={"QUERY_TAG": query_tag},
    )

    cursor = connection.cursor()
    cursor.close()

    return connection


class SnowflakeException(Exception):
    """Exception raised for errors in the Map database connection"""


def is_still_running(map_connection: Connection, query_id: str) -> bool:
    """Check if a Map query is still running."""
    try:
        return map_connection.is_still_running(map_connection.get_query_status_throw_if_error(query_id))
    except Exception as err:
        raise SnowflakeException(err) from err


async def execute_query_async(
    cursor: KomodoSnowflakeCursor,
    query: str,
    query_params: list[str] | dict[str, str] | None = None,
) -> KomodoSnowflakeCursor:
    """Wrapper around the KomodoSnowflakeCursor execute_async method to execute a query 'asynchronously' and poll for completion.

    Note that because the io is still blocking, this function is not truly asynchronous and will get thrown to a sync threadpool
    in FastAPI. If you are working outside of FastAPI, you should probably use `asyncio.to_thread` to run this function.

    Parameters
    ----------
    cursor: an existing KomodoSnowflakeCursor
    query: the query to execute
    query_params: the parameters to pass to the query (optional)

    Example usage:
        connection = get_snowflake_connection(account_id=account_id, jwt=user_jwt)
        cursor = connection.cursor()
        cursor.execute("USE ROLE CUSTOMER_ROLE")
        cursor.execute("USE DATABASE DATA")
        query = "SELECT column_name, table_name FROM INFORMATION_SCHEMA.COLUMNS;"
        cursor = await execute_query_async(cursor=cursor, query=query)
        rows = cursor.fetchall()
    """
    logger.info(f"Executing query:\n{query} with {query_params=}")
    cursor.execute_async(query, query_params)
    query_id = cursor.sfqid
    while is_still_running(cursor.connection, query_id):
        await asyncio.sleep(1)

    cursor.get_results_from_sfqid(query_id)
    return cursor



import logging.config
from uuid import UUID
from typing import Optional

from pydantic import SecretStr, Field
from pydantic_settings import BaseSettings, SettingsConfigDict


logger = logging.getLogger(__name__)


class SDKSettings(BaseSettings):
    # All environment variables related to the komodo platform should start with the prefix "KOMODO_"
    model_config = SettingsConfigDict(env_prefix="KOMODO_")

    environment: str = Field(default="production", env="ENVIRONMENT")
    access_token: Optional[SecretStr] = Field(default=None, env="ACCESS_TOKEN")
    refresh_token: Optional[SecretStr] = Field(default=None, env="REFRESH_TOKEN")
    token_expiration: Optional[int] = Field(default=None, env="TOKEN_EXPIRATION")
    client_id: Optional[str] = Field(default=None, env="CLIENT_ID")
    client_secret: Optional[SecretStr] = Field(default=None, env="CLIENT_SECRET")
    account_id: Optional[UUID] = Field(default=None, env="ACCOUNT_ID")
    account_slug: Optional[str] = Field(default=None, env="ACCOUNT_SLUG")

    @property
    def komodo_external(self) -> str:
        return "https://api.komodohealth.com" if self.environment == "production" else "https://dev-api.khinternal.net"

    @property
    def proxy_domain(self) -> str:
        return "https://connector-gateway.onkomodo.com" if self.environment == "production" else "https://connector-gateway.staging.onkomodo.com"

    @property
    def komodo_rbac_me(self) -> str:
        return f"{self.komodo_external}/platform-rbac-prod/v1/iam/users/me" if self.environment == "production" else f"{self.komodo_external}/platform-rbac-stage/v1/iam/users/me"

    @property
    def komodo_idp_domain(self) -> str:
        return "auth.komodohealth.com" if self.environment == "production" else "auth-staging.komodohealth.com"

    @property
    def komodo_client_id(self) -> str:
        return "mkxXDmAae54idV6TIpviUfq2ySuRdC9Y" if self.environment == "production" else "O69UYSTsC6Tz2jBgzSC30odgX5fmu0xh"

    @property
    def komodo_idp_audience(self) -> str:
        return "https://komodohealth-platform.us.auth0.com/api/v2/" if self.environment == "production" else "https://komodohealth-platform-staging.us.auth0.com/api/v2/"



import asyncio
import os
import threading
from typing import Any, Dict, Tuple

import httpx
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from komodo._vendored.komodo_connector.connection import Connection

HTTP_TOTAL_TIMEOUT = 120
HTTP_CONNECTION_TIMEOUT = 60

timeout = httpx.Timeout(HTTP_TOTAL_TIMEOUT, connect=HTTP_CONNECTION_TIMEOUT)


class SnowflakeRotateKeysCreator:
    """
    Initializes a SnowflakeRotateKeysCreator instance.

    Args:
        driver: The driver used to connect to Snowflake.
        connection_config: The configuration parameters for the Snowflake connection.

    Raises:
        AssertionError: If `komodo_user` is not a string.

    """

    def __init__(self, driver: Any, connection_config: Dict[str, Any]) -> None:
        self.driver = driver
        if not isinstance(connection_config["komodo_user"], str):
            raise AssertionError("komodo_user must be a string")
        self.user = connection_config["komodo_user"].split("+")[1]
        self.connection_config = connection_config

    def fetch_credentials(self) -> Dict[str, str]:
        """
        Fetches the credentials for the Snowflake connection.

        Returns:
            The fetched credentials.

        Raises:
            Exception: If the credentials cannot be fetched.

        """
        url, headers = self.build_connection_url()

        response = httpx.get(
            url,
            params={
                "kh_user": self.user,
            },
            headers=headers,
            timeout=timeout,
            follow_redirects=True,
        )

        if response.status_code != 200:
            raise httpx.HTTPError(
                f"Failed to fetch credentials for user {self.user} with status code {response.status_code}"
            )

        return response.json()  # type: ignore

    def create(self) -> Connection:
        """
        Creates a Snowflake connection using the fetched credentials.

        Returns:
            The Snowflake connection.

        """
        credentials = self.fetch_credentials()
        self.connection_config.update(credentials)

        private_key_str = self.connection_config["USER_PRIVATE_KEY"]

        private_key_obj = serialization.load_pem_private_key(
            private_key_str.encode(), password=None, backend=default_backend()
        )

        private_key_der_bytes = private_key_obj.private_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )

        driver = self.driver(
            user=self.user,
            account=self.connection_config["komodo_account"],
            private_key=private_key_der_bytes,
        )

        if driver.session_id:
            # Reset credentials upon a successful session creation
            self.trigger_reset_credentials()

        return Connection(driver)

    def trigger_reset_credentials(self) -> None:
        """
        Triggers the reset of credentials in a separate thread.

        """

        def run_event_loop(loop: Any) -> None:
            asyncio.set_event_loop(loop)
            loop.run_until_complete(self.reset_credentials())

        # Launch the event loop in a new thread so that it doesn't block, but still runs in the background before the process exits
        loop = asyncio.new_event_loop()
        thread = threading.Thread(target=run_event_loop, args=(loop,))
        thread.start()

    async def reset_credentials(self) -> None:
        """
        Resets the credentials for the Snowflake connection.

        """
        url, headers = self.build_connection_url()
        refresh_credentials_url = f"{url}/refresh"
        async with httpx.AsyncClient(verify=os.getenv("environment") == "prod", timeout=timeout) as client:
            _ = await client.post(
                refresh_credentials_url,
                params={
                    "kh_user": self.user,
                },
                headers=headers,
            )

    def build_connection_url(self) -> Tuple[str, Dict[str, str]]:
        """
        Builds the connection URL and headers for the Snowflake connection.

        Returns:
            The connection URL and headers.

        """
        from komodo._vendored.komodo_connector.setup_driver import get_proxy_host_port

        protocol = self.connection_config.get("query_params", {}).get("protocol", "https")
        host, port = get_proxy_host_port()
        url = f"{protocol}://{host}:{port}/driver/credentials"

        headers = {
            "Authorization": f'Bearer {self.connection_config["komodo_token"]}',
            "Content-Type": "application/json",
            "X-Account-Id": self.connection_config["komodo_account"],
        }
        return url, headers



from __future__ import annotations

import logging
from enum import Enum, unique
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import requests

from komodo._vendored.komodo_connector.connection_creators.snowflake.utils import (
    SNOWFLAKE_VERSION_3_0_0,
    sf_connector_version_greater_than,
)
from komodo._vendored.komodo_connector.setup_driver import Environment, get_proxy_host_port

if sf_connector_version_greater_than(SNOWFLAKE_VERSION_3_0_0):
    from snowflake.connector.auth import AuthByPlugin
else:
    from snowflake.connector.auth_by_plugin import AuthByPlugin

if TYPE_CHECKING:
    from snowflake.connector.network import SnowflakeRestful

KEY_PAIR_AUTHENTICATOR = "SNOWFLAKE_JWT"
logger = logging.getLogger(__name__)


@unique
class KomodoAuthType(Enum):
    KOMODO_PROXY = "KOMODO_PROXY"


class AuthByKomodoProxy(AuthByPlugin):
    """OAuth Based Authentication.

    Works by accepting an OAuth token and using that to authenticate.
    """

    @property
    def type_(self) -> KomodoAuthType:
        return KomodoAuthType.KOMODO_PROXY

    @property
    def assertion_content(self) -> Union[str, None]:
        """Returns the token."""
        return self._token

    def __init__(
        self, kh_token: str, kh_account: str, kh_user: str, komodo_env: str, komodo_host: Optional[str] = None
    ) -> None:
        """Initializes an instance with an OAuth Token."""
        super().__init__()
        self._token: Optional[str] = kh_token
        self._account_id: Optional[str] = kh_account
        self._kh_user: Optional[str] = kh_user
        self._komodo_env: Optional[Environment] = komodo_env
        self._komodo_host: Optional[str] = komodo_host
        self._session = requests.Session()

    def reset_secrets(self) -> None:
        self._token = None
        self._account_id = None
        self._kh_user = None

    def _send_request(self, url: str, params: Dict[str, str], headers: Dict[str, str]) -> requests.Response:
        """Send a GET request and return the response."""
        response = self._session.get(url, params=params, headers=headers)
        response.raise_for_status()
        return response

    def authenticate(self, *args, **kwargs) -> None:  # type: ignore
        """We do not use this method."""
        pass

    def prepare(self, conn: "SnowflakeRestful", **kwargs: Any) -> None:
        """
        Prepare for the authentication process.
        Update the host of the connection with the received host value.
        """
        try:
            conn._rest._host, conn._rest._port = get_proxy_host_port(self._komodo_env, self._komodo_host)
            conn._rest.kh_account = self._account_id
            conn._rest.kh_token = self._token
            conn._rest.kh_user = self._kh_user
            if not sf_connector_version_greater_than(SNOWFLAKE_VERSION_3_0_0):
                conn._rest._token = self._token
        except Exception as e:
            logger.error(f"An error occurred: {e}")

    def reauthenticate(self, **kwargs: Any) -> dict[str, bool]:
        return {"success": False}

    def update_body(self, body: dict[Any, Any]) -> None:
        body["data"]["AUTHENTICATOR"] = KEY_PAIR_AUTHENTICATOR



from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from komodo.iam.models.core.user_lookup_account import UserLookupAccount
from komodo.iam.models.core.user_lookup_group import UserLookupGroup
from typing import Optional, Set, Union
from typing_extensions import Self


class UserLookupOrganization(BaseModel):
    """
    UserLookupOrganization
    """  # noqa: E501

    accounts: Optional[List[UserLookupAccount]] = None
    groups: Optional[List[UserLookupGroup]] = None
    name: StrictStr = Field(description="Organization name")
    organization_id: StrictStr = Field(description="Organization ID")
    slug: StrictStr = Field(description="Organization slug")
    __properties: ClassVar[List[str]] = ["accounts", "groups", "name", "organization_id", "slug"]

    model_config = ConfigDict(populate_by_name=True, validate_assignment=True, protected_namespaces=(), extra="allow")

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of UserLookupOrganization from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in accounts (list)
        _items = []
        if self.accounts:
            for _item_accounts in self.accounts:
                if _item_accounts:
                    _items.append(_item_accounts.to_dict())
            _dict["accounts"] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in groups (list)
        _items = []
        if self.groups:
            for _item_groups in self.groups:
                if _item_groups:
                    _items.append(_item_groups.to_dict())
            _dict["groups"] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of UserLookupOrganization from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "accounts": [UserLookupAccount.from_dict(_item) for _item in obj["accounts"]] if obj.get("accounts") is not None else None,
                "groups": [UserLookupGroup.from_dict(_item) for _item in obj["groups"]] if obj.get("groups") is not None else None,
                "name": obj.get("name"),
                "organization_id": obj.get("organization_id"),
                "slug": obj.get("slug"),
            }
        )
        return _obj



from typing import Dict, Union
from urllib.parse import quote

from pydantic import BaseModel, model_validator

from komodo._vendored.komodo_connector.constants import Environment
from komodo._vendored.komodo_connector.creator import ConnectionCreator


class KomodoAuthenticationParams(BaseModel):
    komodo_account: str
    komodo_user: str
    komodo_token: str

    @model_validator(mode="after")
    def clean_and_quote(self) -> "KomodoAuthenticationParams":
        self.komodo_account = quote(self.komodo_account)
        self.komodo_user = quote(self.komodo_user)

        token = self.komodo_token.strip()

        if token and token[0] in ('"', "'") and token[-1] == token[0]:
            token = token[1:-1]

        # Remove any existing Bearer prefix, ignoring case.
        if token.lower().startswith("bearer "):
            token = token[7:]

        # URL-encode the token and add the standardized Bearer prefix.
        self.komodo_token = f"Bearer {quote(token)}"
        return self


class KomodoConnectionConfig(BaseModel):
    komodo_account: str
    komododb_port: Union[int, None]
    komodo_user: str
    komodo_token: str
    environment: Environment = Environment.PROD
    query_params: Dict[str, Union[str, Dict[str, str]]]

    @model_validator(mode="after")
    def add_bearer(self) -> "KomodoConnectionConfig":
        if not (self.komodo_token.startswith("bearer ") or self.komodo_token.startswith("Bearer ")):
            self.komodo_token = f"Bearer {self.komodo_token}"
        return self


class DriverConfig(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    driver_module_path: str
    driver_creator_class: ConnectionCreator



import io
import re

import httpx
from httpx_retries import RetryTransport, Retry
from komodo.exceptions import ApiException, ApiValueError

SUPPORTED_SOCKS_PROXIES = {"socks5", "socks5h", "socks4", "socks4a"}
RESTResponseType = httpx.Response

ALLOW_RETRY_METHODS = frozenset({"DELETE", "GET", "HEAD", "OPTIONS", "PUT", "TRACE"})


def is_socks_proxy_url(url):
    if url is None:
        return False
    split_section = url.split("://")
    if len(split_section) < 2:
        return False
    else:
        return split_section[0].lower() in SUPPORTED_SOCKS_PROXIES


class RESTResponse(io.IOBase):
    def __init__(self, resp) -> None:
        self.response = resp
        self.status = resp.status_code
        self.reason = resp.reason_phrase
        self.data = None

    def read(self):
        if self.data is None:
            self.data = self.response.content
        return self.data

    async def read_async(self):
        if self.data is None:
            self.data = await self.response.aread()
        return self.data

    def getheaders(self):
        """Returns a dictionary of the response headers."""
        return self.response.headers

    def getheader(self, name, default=None):
        """Returns a given response header."""
        return self.response.headers.get(name, default)


class RESTClientObject:
    def __init__(self, configuration) -> None:
        retry = Retry(total=configuration.retries)
        transport = RetryTransport(retry=retry)
        self.client = httpx.Client(
            transport=transport,
            verify=configuration.verify_ssl,
            timeout=None,
        )

    def request(self, method, url, headers=None, body=None, post_params=None, _request_timeout=None):
        """Perform requests.

        :param method: http request method
        :param url: http request url
        :param headers: http request headers
        :param body: request json body, for `application/json`
        :param post_params: request post parameters,
                            `application/x-www-form-urlencoded`
                            and `multipart/form-data`
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        """
        method = method.upper()
        assert method in ["GET", "HEAD", "DELETE", "POST", "PUT", "PATCH", "OPTIONS"]

        if post_params and body:
            raise ApiValueError("body parameter cannot be used with post_params parameter.")

        post_params = post_params or {}
        headers = headers or {}

        timeout = _request_timeout or 5 * 60

        try:
            # For `POST`, `PUT`, `PATCH`, `OPTIONS`, `DELETE`
            if method in ["POST", "PUT", "PATCH", "OPTIONS", "DELETE"]:
                # no content type provided or payload is json
                content_type = headers.get("Content-Type", "application/json")
                if re.search("json", content_type, re.IGNORECASE):
                    response = self.client.request(method, url, json=body, headers=headers, timeout=timeout)
                elif content_type == "application/x-www-form-urlencoded":
                    response = self.client.request(method, url, data=post_params, headers=headers, timeout=timeout)
                elif content_type == "multipart/form-data":
                    headers.pop("Content-Type", None)
                    response = self.client.request(method, url, files=post_params, headers=headers, timeout=timeout)
                else:
                    response = self.client.request(method, url, content=body, headers=headers, timeout=timeout)
            else:
                response = self.client.request(method, url, headers=headers, timeout=timeout)
        except httpx.RequestError as e:
            raise ApiException(status=0, reason=str(e))

        return RESTResponse(response)


class RESTClientObjectAsync:
    def __init__(self, configuration) -> None:
        retry = Retry(total=configuration.retries)
        transport = RetryTransport(retry=retry)

        self.client = httpx.AsyncClient(
            transport=transport,
            verify=configuration.verify_ssl,
            timeout=None,
        )

    async def close(self) -> None:
        await self.client.aclose()

    async def request(self, method, url, headers=None, body=None, post_params=None, _request_timeout=None):
        """Execute request

        :param method: http request method
        :param url: http request url
        :param headers: http request headers
        :param body: request json body, for `application/json`
        :param post_params: request post parameters,
                            `application/x-www-form-urlencoded`
                            and `multipart/form-data`
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        """
        method = method.upper()
        assert method in ["GET", "HEAD", "DELETE", "POST", "PUT", "PATCH", "OPTIONS"]

        if post_params and body:
            raise ApiValueError("body parameter cannot be used with post_params parameter.")

        post_params = post_params or {}
        headers = headers or {}
        # url already contains the URL query string
        timeout = _request_timeout or 5 * 60

        # Set default Content-Type if not provided
        if "Content-Type" not in headers:
            headers["Content-Type"] = "application/json"

        try:
            # Handle different content types for request body
            if method in ["POST", "PUT", "PATCH", "OPTIONS", "DELETE"]:
                content_type = headers.get("Content-Type", "application/json")
                if re.search("json", content_type, re.IGNORECASE):
                    response = await self.client.request(method, url, json=body, headers=headers, timeout=timeout)
                elif content_type == "application/x-www-form-urlencoded":
                    response = await self.client.request(method, url, data=post_params, headers=headers, timeout=timeout)
                elif content_type == "multipart/form-data":
                    headers.pop("Content-Type", None)
                    response = await self.client.request(method, url, files=post_params, headers=headers, timeout=timeout)
                else:
                    response = await self.client.request(method, url, content=body, headers=headers, timeout=timeout)
            else:
                # Handle GET-like methods with query parameters
                response = await self.client.request(method, url, headers=headers, timeout=timeout)
        except httpx.RequestError as e:
            raise ApiException(status=0, reason=str(e))

        return RESTResponse(response)



# type: ignore
from typing import Dict, Type, Union

from snowflake.snowpark import Session

from komodo._vendored.komodo_connector.connector import connect as komodo_connect


class KomodoSessionBuilder(Session.SessionBuilder):
    def __init__(self) -> None:
        super().__init__()

    def configs(self, options: Dict[str, Union[int, str]]) -> "Session.SessionBuilder":
        conn = komodo_connect(
            account_id=options.get("account_id"),
            profile_id=options.get("profile_id"),
            secret=options.get("secret"),
            port=options.get("port"),
            protocol=options.get("protocol"),
            environment=options.get("environment"),
            session_parameters=options.get("session_parameters"),
        )
        self._options = {
            **self._options,
            **{
                "connection": conn,
            },
        }
        return super().configs(options)


class SnowparkSession(Session):
    @classmethod
    @property
    def builder(cls: Type[Session]) -> "Session.SessionBuilder":
        return KomodoSessionBuilder()

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)



from __future__ import annotations
import json
from enum import Enum
from typing_extensions import Self


class MembershipType(str, Enum):
    """
    MembershipType
    """

    """
    allowed enum values
    """
    MEMBER = "MEMBER"
    ADMIN = "ADMIN"
    NON_DELETE_ADMIN = "NON_DELETE_ADMIN"

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of MembershipType from a JSON string"""
        return cls(json.loads(json_str))



from __future__ import annotations

import functools
import gzip
import json
from typing import TYPE_CHECKING, Any, Dict, Optional, Type

import snowflake
from snowflake.connector.network import NO_TOKEN, AuthBase, SnowflakeRestful

from komodo._vendored.komodo_connector.connection_creators.snowflake.utils import remove_port_from_url
from komodo._vendored.komodo_connector.setup_driver import Environment

HEADER_AUTHORIZATION_KEY = "Authorization"
HEADER_SNOWFLAKE_TOKEN = 'Snowflake Token="{token}"'
KH_ACCOUNT_ID_KEY = "x-account-id"
KH_USER_ID = "x-user-id"
KH_APP_ID = "x-app-id"
KH_SNOWFLAKE_HEADER_AUTHORIZATION_KEY = "Snowflake-Authorization"
COMMON_ERROR_KEYS = ["message", "error", "detail"]
SENSITIVE_STRINGS = ["password", "secret", "token", "key", "credentials"]

if TYPE_CHECKING:
    from snowflake.connector.vendored.requests.sessions import (
        Session as SnowflakeVendorSession,
    )


class KomodoSnowflakeAuth(AuthBase):
    """Attaches HTTP Authorization header for Snowflake."""

    def __init__(self, token: str) -> None:
        # setup any auth-related data here
        self.token = token

    def __call__(self, r: Any) -> Any:
        """Modifies and returns the request."""

        # KH Proxy Support
        if KH_ACCOUNT_ID_KEY in r.headers:
            if self.token != NO_TOKEN:
                r.headers[KH_SNOWFLAKE_HEADER_AUTHORIZATION_KEY] = HEADER_SNOWFLAKE_TOKEN.format(token=self.token)
            return r

        if HEADER_AUTHORIZATION_KEY in r.headers:
            del r.headers[HEADER_AUTHORIZATION_KEY]
        if self.token != NO_TOKEN:
            r.headers[HEADER_AUTHORIZATION_KEY] = HEADER_SNOWFLAKE_TOKEN.format(token=self.token)
        return r


snowflake.connector.network.__dict__["SnowflakeAuth"] = KomodoSnowflakeAuth


class ProxySnowflakeRestful(SnowflakeRestful):
    def __init__(
        self,
        *args: Optional[Any],
        kh_token: Optional[str] = None,
        kh_account: Optional[str] = None,
        kh_user: Optional[str] = None,
        app_id: Optional[str] = None,
        komodo_env: Environment,
        **kwargs: Optional[Any],
    ):
        self.kh_token = kh_token
        self.kh_account = kh_account
        self.kh_user = kh_user
        self.app_id = app_id
        self.komodo_env = komodo_env
        super().__init__(*args, **kwargs)

    def _sanitize_error_message(self, error_body: Any) -> str:
        """
        Sanitizes error messages to prevent leaking sensitive information.

        In order for a proxy service response to be surfaced to the user, it has to be explicitly marked with an X-Komodo-Error header. As such, this is simply an additional fallback in case the proxy service accidentally returns a sensitive response - hence, the naive approach. Update this if a more sophisticated approach is needed.



        Args:
            error_body: Raw error response from the server

        Returns:
            str: Sanitized error message
        """
        try:
            if isinstance(error_body, str) and (parsed := json.loads(error_body)):
                error_body = parsed

            # ruff: noqa: SIM102 (ruff wants to combine the if statements, but that hurts readability in this case)
            if isinstance(error_body, dict):
                if error_msg := next((str(error_body[key]) for key in COMMON_ERROR_KEYS if key in error_body), None):
                    # TODO: Doing a very naive check here. Move to a more sophisticated approach if needed.
                    error_msg_contains_sensitive_info = any(
                        sensitive_string in error_msg.lower().strip() for sensitive_string in SENSITIVE_STRINGS
                    )
                    if not error_msg_contains_sensitive_info:
                        return error_msg

            return "An error occurred while processing your request"
        except json.JSONDecodeError:
            return "An error occurred while processing your request"

    def _check_for_komodo_err(
        self, request: Type[SnowflakeVendorSession.request], *args: Dict[str, Any], **kwargs: Dict[str, Any]
    ) -> Any:
        """
        Checks for Komodo-specific errors and HTTP errors in the response.

        Args:
            request: The request function to execute
            *args: Positional arguments for the request
            **kwargs: Keyword arguments for the request

        Returns:
            Any: The response if no errors are found

        Raises:
            Exception: If a Komodo error or HTTP error is detected
        """
        response = request(*args, **kwargs)
        if komodo_error := getattr(response, "headers", {}).get("x-komodo-error"):
            error_message = f"Komodo Error: {komodo_error}"
            trace_id = self.fetch_trace_id_request(kwargs)
            if trace_id:
                error_message += f" (Trace ID: {trace_id})"
            if (status_code := getattr(response, "status_code")) is not None:
                error_message += f" - Status Code: {status_code}"
            raise Exception(error_message)

        # Check for HTTP errors
        if not (200 <= (status_code := response.status_code) < 300):
            try:
                if (
                    hasattr(response, "json")
                    and (error_body := response.json())
                    or hasattr(response, "text")
                    and (error_body := response.text)
                ):
                    pass
                else:
                    error_body = "Unknown error occurred"
            except (AttributeError, json.JSONDecodeError):
                error_body = "Unknown error occurred"

            sanitized_message = self._sanitize_error_message(error_body)
            error_message = f"HTTP Error {status_code}: {sanitized_message}"
            trace_id = self.fetch_trace_id_request(kwargs)
            if trace_id:
                error_message += f" (Trace ID: {trace_id})"
            raise Exception(error_message)

        return response

    def fetch_trace_id_request(self, kwargs: Dict[str, Any]) -> Optional[str]:
        """
        Extracts trace ID from gzipped request data if available.

        Args:
            kwargs: Request keyword arguments containing data and headers

        Returns:
            Optional[str]: Trace ID if found in the request data, None otherwise
        """
        trace_id = None
        if kwargs.get("data") and kwargs.get("headers", {}).get("Content-Encoding") == "gzip":
            try:
                decompressed_body = gzip.decompress(kwargs["data"])
                body_dict = json.loads(decompressed_body)
                query_tag = body_dict.get("data", {}).get("SESSION_PARAMETERS", {}).get("QUERY_TAG")
                if query_tag:
                    tag_data = json.loads(query_tag)
                    trace_id = tag_data.get("root", {}).get("trace", {}).get("trace_id")
            except (OSError, json.JSONDecodeError):
                pass
        return trace_id

    def _request_exec(self, session: SnowflakeVendorSession, *args: Dict[str, Any], **kwargs: Dict[str, Any]) -> Any:
        # kill all telemetry sending
        # a basic intercept here eliminates the need for a much more intrusive change to the telemetry module
        if "telemetry" in kwargs["full_url"]:
            return {"code": None, "data": "Log Received", "message": None, "success": True}

        updated_headers = kwargs.get("headers", {})
        if self.kh_account:
            updated_headers[KH_ACCOUNT_ID_KEY] = self.kh_account
        if self.kh_token:
            updated_headers[HEADER_AUTHORIZATION_KEY] = self.kh_token
        if self.kh_user:
            updated_headers[KH_USER_ID] = self.kh_user
        if self.app_id:
            updated_headers[KH_APP_ID] = self.app_id

        kwargs["headers"] = updated_headers
        session.request = functools.partial(self._check_for_komodo_err, session.request)
        return super()._request_exec(session, *args, **kwargs)

    def fetch(
        self,
        method: str,
        full_url: str,
        headers: Dict[str, str],
        data: Optional[Any] = None,
        timeout: Optional[Any] = None,
        **kwargs: Optional[Any],
    ) -> Any:
        updated_url = remove_port_from_url(full_url) if self.komodo_env != Environment.LOCAL else full_url
        return super().fetch(method, updated_url, headers, data, timeout, **kwargs)



from __future__ import annotations
import json
from enum import Enum
from typing_extensions import Self


class UserType(str, Enum):
    """
    UserType
    """

    """
    allowed enum values
    """
    INTERNAL = "INTERNAL"
    EXTERNAL = "EXTERNAL"

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of UserType from a JSON string"""
        return cls(json.loads(json_str))



from functools import lru_cache
from importlib import metadata
from typing import Union
from urllib.parse import urlparse, urlunparse

from packaging.version import Version, parse
from snowflake.connector import errorcode, errors

SNOWFLAKE_VERSION_3_0_0 = "3.0.0"
SQLALCHEMY_VERSION_1_4_0 = "1.4.0"


def is_authentication_error(err: Union[errors.ProgrammingError, errors.DatabaseError]) -> bool:
    return err.errno in (
        errorcode.ER_FAILED_TO_CONNECT_TO_DB,
        errorcode.ER_NO_ACCOUNT_NAME,
        errorcode.ER_NO_PASSWORD,
    )


@lru_cache(maxsize=1)
def sf_connector_version_greater_than(version: str) -> bool:
    return parse(metadata.version("snowflake-connector-python")) > Version(version)  # type: ignore


@lru_cache(maxsize=1)
def sql_alchemy_version_greater_than(version: str) -> bool:
    return parse(metadata.version("sqlalchemy")) > Version(version)  # type: ignore


def remove_port_from_url(url: str) -> str:
    parsed_url = urlparse(url)

    new_netloc = parsed_url.hostname

    path_parts = parsed_url.path.split("/")
    if len(path_parts) > 1 and ":" in path_parts[1]:
        path_parts[1] = path_parts[1].split(":")[0]
    new_path = "/".join(path_parts)

    new_url_parts = (parsed_url.scheme, new_netloc, new_path, parsed_url.params, parsed_url.query, parsed_url.fragment)
    return urlunparse(new_url_parts)



# type: ignore
# fmt: off
import logging
import os
from typing import Dict, Optional, Type

from snowflake.connector import connection, errorcode, errors
from snowflake.connector.constants import (
    PARAMETER_AUTOCOMMIT,
    PARAMETER_CLIENT_PREFETCH_THREADS,
    PARAMETER_CLIENT_SESSION_KEEP_ALIVE,
    PARAMETER_CLIENT_SESSION_KEEP_ALIVE_HEARTBEAT_FREQUENCY,
    PARAMETER_CLIENT_VALIDATE_DEFAULT_PARAMETERS,
    PARAMETER_TIMEZONE,
)

from komodo._vendored.komodo_connector.connection_creators.snowflake.cursor import KomodoSnowflakeCursor
from komodo._vendored.komodo_connector.connection_creators.snowflake.komodo_auth import AuthByKomodoProxy
from komodo._vendored.komodo_connector.connection_creators.snowflake.network import ProxySnowflakeRestful
from komodo._vendored.komodo_connector.connection_creators.snowflake.utils import (
    is_authentication_error,
)
from komodo._vendored.komodo_connector.setup_driver import Environment

logger = logging.getLogger(__name__)

class UnsupportedCursorTypeException(Exception):
    pass


class KomodoSnowflakeConnection(connection.SnowflakeConnection):
    KOMODO_AUTH_PARAMS = ("kh_token", "kh_account", "kh_user")

    def __init__(
        self,
        *args,
        environment: Optional[str],
        komodo_host: Optional[str] = None,
        komodo_credentials: Optional[Dict[str, str]] = None,
        komodo_auth_use_env_vars: bool = False,
        **kwargs,
    ) -> None:
        if not komodo_auth_use_env_vars:
            assert komodo_credentials and isinstance(komodo_credentials, dict)
            self.kh_token, self.kh_account, self.kh_user, self.app_id = (
                komodo_credentials.get("kh_token"),
                komodo_credentials.get("kh_account"),
                komodo_credentials.get("kh_user"),
                kwargs.get("app_id"),
            )
        else:
            self.kh_token, self.kh_account, self.kh_user, self.app_id = (
                os.getenv("KH_TOKEN"),
                os.getenv("KH_ACCOUNT"),
                os.getenv("KH_USER"),
                os.getenv("APP_ID"),
            )

        if not self.kh_token or not self.kh_account or not self.kh_user:
            raise errors.ForbiddenError(
                msg=f"Missing required parameters for Komodo Proxy Authenticator: {self.KOMODO_AUTH_PARAMS}",
                errno=errorcode.ER_INVALID_VALUE,
            )
        # Komodo driver sdk should default to prod environment, but allow for override for testing
        self.komodo_env = Environment(environment) if environment else Environment.PROD
        self.komodo_host = komodo_host

        try:
            super().__init__(*args, **kwargs)
        except errors.DatabaseError as err:
            if not is_authentication_error(err):
                raise err
        # The connect below appears to act as some kind of health-check / to send early telemetry data. Either way,
        # it's not necessary for our purposes and can be skipped
        # self.connect(**kwargs)
        self.komodo_credentials = komodo_credentials

    def connect(self, **kwargs) -> None:
        try:
            if kwargs:
                # Not as sketchy as it seems! In restricted environments we can't hit Snowflake's OCSP Location, and
                # our proxy address isn't even in the json that's returned so this removes upto 0.6s of overhead
                # during initial connection
                super()._SnowflakeConnection__config(insecure_mode=True, disable_ocsp_checks=True, **kwargs)
        except errors.ProgrammingError as err:
            if not is_authentication_error(err):
                raise err
        self.__open_connection()

    def __open_connection(self) -> None:
        """Opens a new network connection."""

        # use_numpy and support_negative_year are set later during execution anyway, but are required params :/
        self.converter = self.converter_class(
            use_numpy=True, support_negative_year=True
        )

        # Make sure that a HTTP proxy isn't specified as it causes a different (untested) code-path later
        if self.proxy_host:
            raise NotImplementedError("HTTP Proxy is not supported in Komodo Snowflake Connections")

        self._rest = ProxySnowflakeRestful(
            kh_token=self.kh_token,
            kh_account=self.kh_account,
            kh_user=self.kh_user,
            app_id=self.app_id,
            komodo_env=self.komodo_env,
            host=self.host,
            port=self.port,
            protocol=self._protocol,
            inject_client_pause=self._inject_client_pause,
            connection=self,
        )
        logger.debug("REST API object was created: %s:%s", self.host, self.port)

        # The remainder of this method is cribbed from Snowflake's implementation unless commented

        if self._session_parameters is None:
            self._session_parameters = {}
        if self._autocommit is not None:
            self._session_parameters[PARAMETER_AUTOCOMMIT] = self._autocommit

        if self._timezone is not None:
            self._session_parameters[PARAMETER_TIMEZONE] = self._timezone

        if self._validate_default_parameters:
            # Snowflake will validate the requested database, schema, and warehouse
            self._session_parameters[PARAMETER_CLIENT_VALIDATE_DEFAULT_PARAMETERS] = (
                True
            )

        if self.client_session_keep_alive is not None:
            self._session_parameters[PARAMETER_CLIENT_SESSION_KEEP_ALIVE] = (
                self._client_session_keep_alive
            )

        if self.client_session_keep_alive_heartbeat_frequency is not None:
            self._session_parameters[
                PARAMETER_CLIENT_SESSION_KEEP_ALIVE_HEARTBEAT_FREQUENCY
            ] = self._validate_client_session_keep_alive_heartbeat_frequency()

        if self.client_prefetch_threads:
            self._session_parameters[PARAMETER_CLIENT_PREFETCH_THREADS] = (
                self._validate_client_prefetch_threads()
            )

        # Explicitly call our auth method instead of relying on a failure path from the 'regular' one.
        self.auth_class = AuthByKomodoProxy(
            self.kh_token, self.kh_account, self.kh_user, self.komodo_env, self.komodo_host
        )
        self.authenticate_with_retry(self.auth_class)

        self._password = None  # ensure password won't persist
        self.auth_class.reset_secrets()

        self.initialize_query_context_cache()

        if self.client_session_keep_alive:
            # This will be called after the heartbeat frequency has actually been set.
            # By this point it should have been decided if the heartbeat has to be enabled
            # and what would the heartbeat frequency be
            self._add_heartbeat()

    def cursor(self, cursor_class: Type[KomodoSnowflakeCursor] = KomodoSnowflakeCursor):
        if not issubclass(cursor_class, KomodoSnowflakeCursor):
            raise UnsupportedCursorTypeException(f"Unsupported cursor type: {cursor_class}")
        return cursor_class(self)

    def execute_string(self, sql_text: str, remove_comments: bool = False, return_cursors: bool = True, cursor_class: Type[KomodoSnowflakeCursor] = KomodoSnowflakeCursor, **kwargs):
        return super().execute_string(sql_text, remove_comments, return_cursors, cursor_class, **kwargs)



from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, SecretStr, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from komodo.iam.models.core.membership_type import MembershipType
from komodo.iam.models.core.role_response import RoleResponse
from typing import Optional, Set, Union
from typing_extensions import Self


class ServicePrincipalResponse(BaseModel):
    """
    ServicePrincipalResponse
    """  # noqa: E501

    account_id: Optional[StrictStr] = Field(default=None, description="Account ID")
    application_id: Optional[StrictStr] = Field(default=None, description="Application ID")
    client_id: StrictStr = Field(description="Client ID")
    client_secret: Optional[SecretStr] = Field(default=None, description="Client secret")
    created_at: datetime = Field(description="Time of creation")
    created_by: Optional[StrictStr] = Field(default=None, description="Created By")
    description: StrictStr = Field(description="ServicePrincipal description")
    membership_type: Optional[Union[MembershipType, str]] = Field(default=None, description="The service principal's level of access within platform-rbac")
    name: StrictStr = Field(description="ServicePrincipal name")
    roles: Optional[List[RoleResponse]] = None
    service_principal_id: StrictStr = Field(description="Service Principal ID")
    updated_at: datetime = Field(description="Time of last update")
    updated_by: Optional[StrictStr] = Field(default=None, description="Updated By")
    __properties: ClassVar[List[str]] = [
        "account_id",
        "application_id",
        "client_id",
        "client_secret",
        "created_at",
        "created_by",
        "description",
        "membership_type",
        "name",
        "roles",
        "service_principal_id",
        "updated_at",
        "updated_by",
    ]

    model_config = ConfigDict(populate_by_name=True, validate_assignment=True, protected_namespaces=(), extra="allow")

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of ServicePrincipalResponse from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in roles (list)
        _items = []
        if self.roles:
            for _item_roles in self.roles:
                if _item_roles:
                    _items.append(_item_roles.to_dict())
            _dict["roles"] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of ServicePrincipalResponse from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "account_id": obj.get("account_id"),
                "application_id": obj.get("application_id"),
                "client_id": obj.get("client_id"),
                "client_secret": obj.get("client_secret"),
                "created_at": obj.get("created_at"),
                "created_by": obj.get("created_by"),
                "description": obj.get("description"),
                "membership_type": obj.get("membership_type"),
                "name": obj.get("name"),
                "roles": [RoleResponse.from_dict(_item) for _item in obj["roles"]] if obj.get("roles") is not None else None,
                "service_principal_id": obj.get("service_principal_id"),
                "updated_at": obj.get("updated_at"),
                "updated_by": obj.get("updated_by"),
            }
        )
        return _obj



from typing import Any, Dict, Optional


class Connection(object):
    """
    Komodo database driver following DB 2.0 API
    This driver is a wrapper around snowflake or databricks drivers
    It is used to connect to the database and execute queries
    It is also used to get the list of tables and columns in the database
    It should completely implement all DB 2.0 API methods for Connection.

    """

    def __init__(self, connection: Any) -> None:
        self.connection = connection

    def close(self, *args: Optional[Any], **kwargs: Optional[Dict[str, Any]]) -> None:
        self.connection.close(*args, **kwargs)

    def commit(self) -> None:
        self.connection.commit()

    def rollback(self) -> None:
        self.connection.rollback()

    def cursor(self, *args: Optional[Any], **kwargs: Optional[Dict[str, Any]]) -> Any:
        return self.connection.cursor(*args, **kwargs)

    def __enter__(self) -> Any:
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.close()

    def __del__(self) -> None:
        del self.connection

    def __getattr__(self, name: str) -> Any:
        return getattr(self.connection, name)



from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from typing import Optional, Set, Union
from typing_extensions import Self


class RoleResponse(BaseModel):
    """
    RoleResponse
    """  # noqa: E501

    application_id: StrictStr = Field(description="Application ID")
    created_at: datetime = Field(description="Time of creation")
    created_by: Optional[StrictStr] = Field(default=None, description="Created By")
    description: StrictStr = Field(description="Role description")
    name: StrictStr = Field(description="Role name")
    organization_id: Optional[StrictStr] = Field(default=None, description="Organization id")
    role_id: StrictStr = Field(description="Role ID")
    updated_at: datetime = Field(description="Time of last update")
    updated_by: Optional[StrictStr] = Field(default=None, description="Updated By")
    __properties: ClassVar[List[str]] = ["application_id", "created_at", "created_by", "description", "name", "organization_id", "role_id", "updated_at", "updated_by"]

    model_config = ConfigDict(populate_by_name=True, validate_assignment=True, protected_namespaces=(), extra="allow")

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of RoleResponse from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of RoleResponse from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "application_id": obj.get("application_id"),
                "created_at": obj.get("created_at"),
                "created_by": obj.get("created_by"),
                "description": obj.get("description"),
                "name": obj.get("name"),
                "organization_id": obj.get("organization_id"),
                "role_id": obj.get("role_id"),
                "updated_at": obj.get("updated_at"),
                "updated_by": obj.get("updated_by"),
            }
        )
        return _obj



from importlib.metadata import version
import os
import typer

from .login import login, jwt
from .account import account_app
from .auxiliary import marmot, docs
from .service_principal import service_principal_app
from .feedback import feedback

__version__ = version("komodo")

app = typer.Typer(
    name="komodo",
    help="Komodo CLI - Build the evidentiary standard",
    add_completion=False,
    invoke_without_command=True,
)


@app.callback()
def callback(ctx: typer.Context, version: bool = typer.Option(False, "--version", "-v", help="Show the version and exit.")):
    """
    Komodo CLI - A command-line interface for managing Komodo workspaces and applications.

    Use this CLI to:
    • Login and manage authentication
    • Set and get account information
    • Build services using BuildKit
    • Manage Komodo applications
    """
    if version:
        typer.echo(f"Komodo CLI Version: {__version__}")
        raise typer.Exit()
    if ctx.invoked_subcommand is None:
        typer.echo(ctx.get_help())
        raise typer.Exit()


app.command()(login)
app.command()(jwt)
app.command()(marmot)
app.command()(docs)
app.command()(feedback)
app.add_typer(account_app, name="account")
app.add_typer(service_principal_app, name="service-principal")

if __name__ == "__main__":
    app()



from komodo.iam.models.core.service_principal_response import ServicePrincipalResponse
from komodo.iam.models.core.user_lookup_account import UserLookupAccount
from komodo.iam.models.core.user_lookup_group import UserLookupGroup
from komodo.iam.models.core.user_lookup_organization import UserLookupOrganization
from komodo.iam.models.core.user_lookup_response import UserLookupResponse
from komodo.iam.models.core.user_status import UserStatus
from komodo.iam.models.core.user_type import UserType
from komodo.iam.models.identity.identity_response import IdentityResponse
from komodo.iam.models.identity.jwt_type import JWTType
from komodo.iam.models.service_principals.service_principal_token_create import ServicePrincipalTokenCreate
from komodo.iam.models.service_principals.service_principal_token_response import ServicePrincipalTokenResponse
from komodo.iam.models.service_principals.service_principal_update import ServicePrincipalUpdate



import os
import typer

from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.text import Text

from cli.util import ENVIRONMENT_OPTION, get_headers, make_api_request, print_table, resolve_environment

service_principal_app = typer.Typer(short_help="Manage Service Principals", no_args_is_help=True)


def get_proxy_url(environment: str) -> str:
    if environment.lower() == "integration":
        return "https://connector-gateway.staging.onkomodo.com/internal/service-principals"
    else:
        return "https://connector-gateway.onkomodo.com/internal/service-principals"


@service_principal_app.command("create")
@resolve_environment
def create_service_principal(
    environment: str = ENVIRONMENT_OPTION,
    name: str = typer.Option(..., "--name", "-n", help="The name of the service principal"),
    description: str = typer.Option(..., "--description", "-d", help="The description of the service principal"),
):
    """Create a new service principal."""
    from komodo.auth.Session import Session

    session = Session(environment=environment)
    credentials = session.load_credentials()
    console = Console()
    if not credentials.access_token:
        console.print(Text("No access token found. Please login first.", style="bold red"))
        raise typer.Exit(code=1)
    if not credentials.account_id:
        console.print(Text("No account ID found. Please run 'komodo account set' first.", style="bold red"))
        raise typer.Exit(code=1)
    with Progress(SpinnerColumn(), TextColumn("[progress.description]{task.description}"), transient=True) as progress:
        task = progress.add_task("Creating service principal...", start=True)
        try:
            url = get_proxy_url(environment)
            headers = get_headers(credentials.access_token, credentials.account_id)
            response = make_api_request(
                "POST",
                url,
                headers=headers,
                json_data={"name": name, "description": description},
                console=console,
                custom_error_messages={403: "Access denied. Only account admins can create service principals."},
            )
        except Exception as e:
            progress.stop()
            console.print(Text(f"Failed to create service principal: {e}", style="bold red"))
            raise typer.Exit(code=1)
        progress.update(task, description="Service principal created!")
    console.print(
        Text(
            f"Service principal {name} created successfully. Client ID and Client Secret are sensitive credentials and should be stored in a secure location. Client ID: {response.get('client_id')} Client Secret: {response.get('client_secret')}"
        )
    )


@service_principal_app.command("list")
@resolve_environment
def list_service_principals(environment: str = ENVIRONMENT_OPTION):
    """List all service principals."""
    from komodo.auth.Session import Session

    session = Session(environment=environment)
    credentials = session.load_credentials()
    console = Console()
    if not credentials.access_token:
        console.print(Text("No access token found. Please login first.", style="bold red"))
        raise typer.Exit(code=1)
    if not credentials.account_id:
        console.print(Text("No account ID found. Please run 'komodo account set' first.", style="bold red"))
        raise typer.Exit(code=1)
    with Progress(SpinnerColumn(), TextColumn("[progress.description]{task.description}"), transient=True) as progress:
        task = progress.add_task("Fetching service principals...", start=True)
        try:
            url = get_proxy_url(environment)
            headers = get_headers(credentials.access_token, credentials.account_id)
            response = make_api_request("GET", url, headers=headers, console=console)
            service_principals = response.get("service_principals", [])
        except Exception as e:
            progress.stop()
            console.print(Text(f"Failed to fetch service principals: {e}", style="bold red"))
            raise typer.Exit(code=1)
        progress.update(task, description="Service principals fetched!")
    print_table(service_principals, "Komodo Service Principals", ["service_principal_id", "name", "description", "created_at"], console=console)


@service_principal_app.command("delete")
@resolve_environment
def delete_service_principal(environment: str = ENVIRONMENT_OPTION, service_principal_id: str = typer.Option(..., "--service-principal-id", "-s", help="The ID of the service principal to delete")):
    """Delete a service principal."""
    from komodo.auth.Session import Session

    session = Session(environment=environment)
    credentials = session.load_credentials()
    console = Console()
    if not credentials.access_token:
        console.print(Text("No access token found. Please login first.", style="bold red"))
        raise typer.Exit(code=1)
    if not credentials.account_id:
        console.print(Text("No account ID found. Please run 'komodo account set' first.", style="bold red"))
        raise typer.Exit(code=1)
    with Progress(SpinnerColumn(), TextColumn("[progress.description]{task.description}"), transient=True) as progress:
        task = progress.add_task("Deleting service principal...", start=True)
        try:
            url = get_proxy_url(environment) + f"/{service_principal_id}"
            headers = get_headers(credentials.access_token, credentials.account_id)
            make_api_request("DELETE", url, headers=headers, console=console, custom_error_messages={403: "Access denied. Only account admins can delete service principals."})
        except Exception as e:
            progress.stop()
            console.print(Text(f"Failed to delete service principal: {e}", style="bold red"))
            raise typer.Exit(code=1)
        progress.update(task, description="Service principal deleted!")
    console.print(Text(f"Service principal {service_principal_id} deleted successfully.", style="bold green"))



import logging

import httpx
import questionary
import typer
from rich.console import Console
from rich.text import Text

from komodo.config import SDKSettings
from komodo.auth.Session import Session
from cli.util import resolve_environment, ENVIRONMENT_OPTION

logger = logging.getLogger(__name__)

@resolve_environment
def feedback(environment: str = ENVIRONMENT_OPTION):
    """Submit feedback about the Komodo Connector."""
    console = Console()

    # Get JWT token
    settings = SDKSettings(environment=environment)
    session = Session(environment=settings.environment)
    credentials = session.load_credentials()
    
    if not credentials or not credentials.access_token:
        console.print(Text("No access token found. Please run 'komodo login' first.", style="bold red"))
        raise typer.Exit(code=1)
    
    jwt_token = credentials.access_token

    try:
        headers = {"Authorization": f"Bearer {jwt_token}", "accept": "application/json"}
        
        with httpx.Client() as client:
            response = client.get(settings.komodo_rbac_me, headers=headers, timeout=30.0)
            response.raise_for_status()
            user_data = response.json()
            email = user_data.get("email")
            
            if not email:
                console.print(Text("Can't retrieve email of authenticated user", style="bold red"))
                raise typer.Exit(code=1)
    except typer.Exit:
        raise
    except Exception as e:
        logger.debug(f"Failed to get email from RBAC API: {e}", exc_info=True)
        console.print(Text("Can't retrieve email of authenticated user", style="bold red"))
        raise typer.Exit(code=1) from e

    # Prompt for feedback
    feedback_text = questionary.text("Enter your feedback or bug report:", multiline=True).ask()
    if not feedback_text:
        console.print(Text("Feedback is required. Exiting.", style="bold red"))
        raise typer.Exit(code=1)

    url = f"{settings.proxy_domain}/feedback"
    payload = {"email": email, "body": feedback_text}
    headers = {"Authorization": f"Bearer {jwt_token}", "Content-Type": "application/json"}

    # Send POST request
    try:
        with httpx.Client() as client:
            response = client.post(url, json=payload, headers=headers, timeout=30.0)
            response.raise_for_status()
            console.print(Text("Your feedback has been sent!", style="bold green"))
    except httpx.HTTPStatusError as e:
        console.print(Text("Failed to send feedback. Please try again later.", style="bold red"))
        logger.debug(f"Server response: {e.response.status_code} - {e.response.text}")
        raise typer.Exit(code=1) from e
    except httpx.RequestError as e:
        console.print(Text("Failed to send feedback. Please check your connection and try again.", style="bold red"))
        logger.debug(f"Request error: {e}")
        raise typer.Exit(code=1) from e
    except Exception as e:
        console.print(Text("An unexpected error occurred while sending feedback.", style="bold red"))
        logger.debug(f"Unexpected error: {e}")
        raise typer.Exit(code=1) from e




"""
This module provides pydantic classes to build a standard Snowflake query tag for KH applications and services.

Example usage:
>>> env = EnvEnum("prod")
>>> type = "eng"
>>> nest = "compute"
>>> account_id = "00000000-0000-0000-0000-000000000000"
>>> user_id = "00000000-0000-0000-0000-000000000000"
>>> service = VersionedProperties(name="ucc",version="1.0.0", properties={"cohort_run_id": 42})
>>> client = VersionedProperties(name="prism",version="1.0.0", properties={})
>>> trace = Trace(trace_id="00000000-0000-0000-0000-000000000000", parent_id="00000000-0000-0000-0000-000000000000")
>>> snowflakeTag = SnowflakeTag(type=type,env=env, nest=nest, account_id=account_id, user_id=user_id, service=service, client=client,trace=trace)
>>> query_tag = snowflakeTag.model_dump_json(exclude_none=True,exclude_unset=True, by_alias=True, indent=4)

Output (formatted):
{
    "env": "prod",
    "nest": "compute",
    "account_id": "00000000-0000-0000-0000-000000000000",
    "type": "eng",
    "user_id": "00000000-0000-0000-0000-000000000000",
    "trace": {
        "trace_id": "00000000-0000-0000-0000-000000000000",
        "parent_id": "00000000-0000-0000-0000-000000000000"
    },
    "service": {
        "name": "ucc",
        "properties": {
            "cohort_run_id": 42
        },
        "version": "1.0.0"
    },
    "client": {
        "name": "prism",
        "properties": {},
        "version": "1.0.0"
    }
}
"""


__version__ = "1.0.0"



# type: ignore
# fmt: off
import logging
import re
import traceback

from snowflake.connector import cursor

logger = logging.getLogger(__name__)


class KomodoSnowflakeCursor(cursor.SnowflakeCursor):
    USE_WAREHOUSE_QUERY = r"\buse\s+warehouse\b"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def execute(self, command: str, *args, **kwargs) -> None:
        if re.search(self.USE_WAREHOUSE_QUERY, command, re.IGNORECASE):
            logging.warning(f"Invalid Query: {command}")
            raise ValueError("Invalid Query: The 'USE WAREHOUSE' command is not permitted.")
        try:
            return super().execute(command, *args, **kwargs)
        except AttributeError as e:
            if "serialize_to_dict" in str(e):
                logger.error(
                    f"Error executing query: {command}. Error: {e}. Traceback: {traceback.format_exc()}"
                )
                raise Exception(
                    "This query failed to execute. Please check the query syntax or connection settings or permissions on the role used.")
            raise


class KomodoSnowflakeDictCursor(KomodoSnowflakeCursor):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs, use_dict_result=True)



import os
from pydantic import validate_call, Field, StrictFloat, StrictStr, StrictInt
from typing import Any, Dict, List, Optional, Tuple, Union
from typing_extensions import Annotated

from pydantic import Field, StrictStr
from typing import Any, List, Optional
from typing_extensions import Annotated
from komodo.iam.models.identity.identity_response import IdentityResponse
from komodo.api_client import ApiClient, RequestSerialized


class IamApi:
    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client
        self.api_version = "2024-10-01"

    @validate_call
    def get_current_identity(
        self,
        expand: Annotated[
            Optional[List[StrictStr]],
            Field(
                description='Include expanded resources related to the user in the response:                - `organizations`: include a list of all organizations to which the user belongs                - `organizations.accounts`: a list of all organizations to which the user belongs _with                     a nested list of the accounts to which the user belongs within each organization_                - `organizations.groups`: a list of all organizations to which the user belongs _with a nested                     list of the groups to which the user belongs within each organization_                - `organizations.groups.accounts`: a list of all organizations to which the user belongs _with                     a nested list of the accounts to which the user has direct membership                     within each organization, and a nested list of the groups to which the                     user belongs within the organization. Each group contains a nested list of accounts to which                    the group (and therefore the user) has membership in the organization_               Note:                - If multiple elements with the same prefix but _different levels of nesting_ are specified                     only the longest element will be used. For example,                     if `"expand" = ["organizations", "organizations.accounts"]` the result                     will be the same as if `"expand" = ["organizations.accounts"]`.                - If both `organizations.accounts` and `organizations.groups.accounts` is                     specified, the behavior will match `organizations.groups.accounts`                - Elements must be unique             '
            ),
        ] = None,
        _request_timeout: Union[None, Annotated[StrictFloat, Field(gt=0)], Tuple[Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=2)] = 0,
    ) -> IdentityResponse:
        """Look up details on the User or Service Principal making the request.

        Look up details on the User or Service Principal making the request.  Look up details on the User or Service Principal making the request. The 'expand' parameter allows additional information to be returned for Users.  **Returns:** - IdentityResponse: Details of the User or Service Principal making the request.

        :param expand: Include expanded resources related to the user in the response:                - `organizations`: include a list of all organizations to which the user belongs                - `organizations.accounts`: a list of all organizations to which the user belongs _with                     a nested list of the accounts to which the user belongs within each organization_                - `organizations.groups`: a list of all organizations to which the user belongs _with a nested                     list of the groups to which the user belongs within each organization_                - `organizations.groups.accounts`: a list of all organizations to which the user belongs _with                     a nested list of the accounts to which the user has direct membership                     within each organization, and a nested list of the groups to which the                     user belongs within the organization. Each group contains a nested list of accounts to which                    the group (and therefore the user) has membership in the organization_               Note:                - If multiple elements with the same prefix but _different levels of nesting_ are specified                     only the longest element will be used. For example,                     if `\"expand\" = [\"organizations\", \"organizations.accounts\"]` the result                     will be the same as if `\"expand\" = [\"organizations.accounts\"]`.                - If both `organizations.accounts` and `organizations.groups.accounts` is                     specified, the behavior will match `organizations.groups.accounts`                - Elements must be unique
        :type expand: List[str]
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """  # noqa: E501

        _param = self._get_current_identity_serialize(expand=expand, _request_auth=_request_auth, _content_type=_content_type, _headers=_headers, _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            "200": "IdentityResponse",
            "404": "ErrorResponse",
            "422": "HTTPValidationError",
        }

        response_data = self.api_client.call_api(*_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            subpackage="komodo.iam.models",
            response_types_map=_response_types_map,
        ).data

    @validate_call
    async def get_current_identity_async(
        self,
        expand: Annotated[
            Optional[List[StrictStr]],
            Field(
                description='Include expanded resources related to the user in the response:                - `organizations`: include a list of all organizations to which the user belongs                - `organizations.accounts`: a list of all organizations to which the user belongs _with                     a nested list of the accounts to which the user belongs within each organization_                - `organizations.groups`: a list of all organizations to which the user belongs _with a nested                     list of the groups to which the user belongs within each organization_                - `organizations.groups.accounts`: a list of all organizations to which the user belongs _with                     a nested list of the accounts to which the user has direct membership                     within each organization, and a nested list of the groups to which the                     user belongs within the organization. Each group contains a nested list of accounts to which                    the group (and therefore the user) has membership in the organization_               Note:                - If multiple elements with the same prefix but _different levels of nesting_ are specified                     only the longest element will be used. For example,                     if `"expand" = ["organizations", "organizations.accounts"]` the result                     will be the same as if `"expand" = ["organizations.accounts"]`.                - If both `organizations.accounts` and `organizations.groups.accounts` is                     specified, the behavior will match `organizations.groups.accounts`                - Elements must be unique             '
            ),
        ] = None,
        _request_timeout: Union[None, Annotated[StrictFloat, Field(gt=0)], Tuple[Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=2)] = 0,
    ) -> IdentityResponse:
        """Look up details on the User or Service Principal making the request.

        Look up details on the User or Service Principal making the request.  Look up details on the User or Service Principal making the request. The 'expand' parameter allows additional information to be returned for Users.  **Returns:** - IdentityResponse: Details of the User or Service Principal making the request.

        :param expand: Include expanded resources related to the user in the response:                - `organizations`: include a list of all organizations to which the user belongs                - `organizations.accounts`: a list of all organizations to which the user belongs _with                     a nested list of the accounts to which the user belongs within each organization_                - `organizations.groups`: a list of all organizations to which the user belongs _with a nested                     list of the groups to which the user belongs within each organization_                - `organizations.groups.accounts`: a list of all organizations to which the user belongs _with                     a nested list of the accounts to which the user has direct membership                     within each organization, and a nested list of the groups to which the                     user belongs within the organization. Each group contains a nested list of accounts to which                    the group (and therefore the user) has membership in the organization_               Note:                - If multiple elements with the same prefix but _different levels of nesting_ are specified                     only the longest element will be used. For example,                     if `\"expand\" = [\"organizations\", \"organizations.accounts\"]` the result                     will be the same as if `\"expand\" = [\"organizations.accounts\"]`.                - If both `organizations.accounts` and `organizations.groups.accounts` is                     specified, the behavior will match `organizations.groups.accounts`                - Elements must be unique
        :type expand: List[str]
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """  # noqa: E501

        _param = self._get_current_identity_serialize(expand=expand, _request_auth=_request_auth, _content_type=_content_type, _headers=_headers, _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            "200": "IdentityResponse",
            "404": "ErrorResponse",
            "422": "HTTPValidationError",
        }
        response_data = await self.api_client.call_api_async(*_param, _request_timeout=_request_timeout)
        await response_data.read_async()
        return self.api_client.response_deserialize(
            response_data=response_data,
            subpackage="komodo.iam.models",
            response_types_map=_response_types_map,
        ).data

    def _get_current_identity_serialize(
        self,
        expand,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:
        _hosts = ["https://dev-api.khinternal.net", "https://api.komodohealth.com"]
        _host = _hosts[_host_index]

        environment = os.getenv("KOMODO_ENVIRONMENT", "production")
        if environment == "integration":
            _host = _hosts[0]
        else:
            _host = _hosts[1]

        _collection_formats: Dict[str, str] = {
            "expand": "multi",
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if expand is not None:
            _query_params.append(("expand", expand))

        # process the header parameters
        # process the form parameters
        # process the body parameter

        # set the HTTP header `Accept`
        if "Accept" not in _header_params:
            _header_params["Accept"] = self.api_client.select_header_accept(["application/json"])

        _header_params["X-API-Version"] = self.api_version
        # authentication setting
        _auth_settings: List[str] = ["HTTPBearer"]

        return self.api_client.param_serialize(
            method="GET",
            resource_path="/v1/iam/identity",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth,
        )



class APIResponseError(Exception):
    def __init__(self, message: str, status_code: int, payload: str):
        self.message = message
        self.status_code = status_code
        self.payload = payload
        super().__init__(self.message)

    def __str__(self):
        return f"{self.message} [{self.status_code}]. Payload={self.payload}"



from __future__ import annotations
import json
from enum import Enum
from typing_extensions import Self


class UserStatus(str, Enum):
    """
    UserStatus
    """

    """
    allowed enum values
    """
    INACTIVE = "inactive"
    ACTIVE = "active"

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of UserStatus from a JSON string"""
        return cls(json.loads(json_str))



from typing import TYPE_CHECKING, Any, Dict, Union

import httpx

from komodo._vendored.komodo_connector.connection import Connection

HTTP_TOTAL_TIMEOUT = 120
HTTP_CONNECTION_TIMEOUT = 60

timeout = httpx.Timeout(HTTP_TOTAL_TIMEOUT, connect=HTTP_CONNECTION_TIMEOUT)
if TYPE_CHECKING:
    from sqlalchemy.engine import URL


class SnowflakeNativeCreator:
    def __init__(self, driver: Any, connection_config: Dict[str, Any]) -> None:
        """
        Initializes a SnowflakeNativeCreator instance.

        Args:
            driver: The driver used to connect to Snowflake.
            connection_config: The configuration parameters for the Snowflake connection.

        """
        self.driver = driver
        self.connection_config = {
            "user": connection_config["komodo_user"],
            "password": connection_config["komodo_token"],
            "account": connection_config["komodo_account"],
            "session_parameters": connection_config.get("query_params", {}).get("session_parameters", {}),
        }

    def create(self) -> Connection:
        """
        Creates a Snowflake connection using the provided configuration parameters.

        Returns:
            The Snowflake connection.

        """
        return Connection(self.driver(**self.connection_config))

    def build_connection_url(self) -> Union[str, "URL", None]: ...



from functools import singledispatch
from importlib import import_module
from typing import Dict, Optional, Tuple, Union
from urllib.parse import parse_qsl, urlparse

from sqlalchemy.engine.url import URL

from komodo._vendored.komodo_connector.connection_creators.snowflake.creator import SnowflakeCreator
from komodo._vendored.komodo_connector.connection_creators.snowflake_native.creator import (
    SnowflakeNativeCreator,
)
from komodo._vendored.komodo_connector.connection_creators.snowflake_rotate_keys.creator import (
    SnowflakeRotateKeysCreator,
)
from komodo._vendored.komodo_connector.constants import DEV_KOMODO_PROXY_SERVICE_DEFAULT_PORT, KOMODO_PROXY_HOST, Environment
from komodo._vendored.komodo_connector.creator import ConnectionCreator
from komodo._vendored.komodo_connector.models import (
    DriverConfig,
    KomodoConnectionConfig,
)

# Do not import the driver classes in this module here to avoid the memory overhead of loading all the drivers
# We will dynamically load the required driver classes in the load_driver function


DRIVERS_MAP: Dict[str, DriverConfig] = {
    # SnowflakeCreator - routes all connections through the proxy
    "snowflake": DriverConfig(
        driver_module_path="komodo._vendored.komodo_connector.connection_creators.snowflake.connect.KomodoSnowflakeConnection",
        driver_creator_class=SnowflakeCreator,
    ),
    # SnowflakeRotateKeysCreator - routes all connections directly to snowflake,
    # but fetches temporary snowflake credentials from the proxy and rotates them
    "snowflake-keys": DriverConfig(
        driver_module_path="snowflake.connector.connect",
        driver_creator_class=SnowflakeRotateKeysCreator,
    ),
    # SnowflakeNativeCreator - routes all connections directly to snowflake
    "snowflake-native": DriverConfig(
        driver_module_path="snowflake.connector.connect",
        driver_creator_class=SnowflakeNativeCreator,
    ),
}


@singledispatch
def parse_komodo_url(url: str) -> Tuple[str, KomodoConnectionConfig]:
    """
    The `parse_url` function takes a URL as input and returns a tuple containing the vendor name and a
    dictionary of parsed components from the URL.

    :param url: The `url` parameter is a string that represents a URL
    :return: The function `parse_url` returns a tuple containing two elements. The first element is a
    string representing the vendor, and the second element is a dictionary containing the
    connection parameters
    """
    parsed_url = urlparse(url)
    vendor = parsed_url.netloc.split("+", 1)[0]

    user_auth, account_config = parsed_url.netloc.split("@", 1)
    komodo_user_string, komodo_token = user_auth.split(":")[:2]
    komodo_user = komodo_user_string.split("+", 1)[1]
    komodo_account, komododb_port = (
        account_config.split(":", 1)
        if ":" in account_config
        else (account_config, DEV_KOMODO_PROXY_SERVICE_DEFAULT_PORT)
    )

    query_param_dict: Dict[str, Union[str, Dict[str, str]]] = dict(parse_qsl(parsed_url.query))

    return (
        vendor,
        KomodoConnectionConfig(
            komodo_account=komodo_account,
            komododb_port=int(komododb_port),
            komodo_user=komodo_user,
            komodo_token=komodo_token,
            query_params=query_param_dict,
        ),
    )


@parse_komodo_url.register
def parse_url_sqlalchemy_URL(url: URL) -> Tuple[str, KomodoConnectionConfig]:
    """
    The function `parse_url_sqlalchemy_URL` takes a URL object and returns a tuple containing the host,
    database, username, password, and port extracted from the URL.
    """
    vendor, komodo_user_string = url.username.split("+", 1)
    komodo_user = komodo_user_string.split("+", 1)[1]
    komodo_port = url.port or DEV_KOMODO_PROXY_SERVICE_DEFAULT_PORT
    return (
        vendor,
        KomodoConnectionConfig(
            komodo_account=url.host,
            komododb_port=int(komodo_port),
            komodo_user=komodo_user,
            komodo_token=url.password,
            query_params=url.query,
        ),
    )


def load_driver(driver_name: str) -> Tuple[object, ConnectionCreator]:
    driver_config = DRIVERS_MAP[driver_name]
    module_path, driver_name = driver_config.driver_module_path.rsplit(".", 1)

    driver_module = import_module(module_path)
    # TODO: Support multiple driver versions by reloading the driver module with the specified version
    # importlib.reload(driver_module)

    driver = getattr(driver_module, driver_name)
    return driver, driver_config.driver_creator_class


def get_proxy_host_port(
    komodo_env: Environment = Environment.PROD, komodo_host: Optional[str] = None
) -> Tuple[str, str]:
    host = komodo_host if (komodo_env != Environment.PROD and komodo_host) else KOMODO_PROXY_HOST[komodo_env]
    port = DEV_KOMODO_PROXY_SERVICE_DEFAULT_PORT
    return host, port



import configparser
import logging
import threading
import time
import os

from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Optional

from komodo.auth.constants import KOMODO_IDP_API_SCOPE
from komodo.auth.OAuthFlows import OAuthFlows
from komodo.config import SDKSettings
from pydantic import BaseModel

logger = logging.getLogger(__name__)


class ExpiredCredentialsError(Exception):
    def __init__(self, message: str = "Credentials are expired and need to be refreshed."):
        super().__init__(message)
        self.message = message


class AuthenticationType(Enum):
    PASS_THROUGH = "PassThrough"
    BROWSER = "Browser"
    MACHINE_TO_MACHINE = "MachineToMachine"


class Credentials(BaseModel):
    access_token: str | None = None
    expires_at: int | None = None
    client_id: str | None = None
    client_secret: str | None = None
    account_id: str | None = None
    account_slug: str | None = None
    credential_type: AuthenticationType | None = None

    def __init__(self, **data):
        logging.debug(f"Credentials initialized with keys: {list(data.keys())}")
        super().__init__(**data)

    def refresh_needed(self) -> bool:
        """
        Check if the access token needs to be refreshed within the next 15 minutes.
        :return: True if refresh is needed, False otherwise
        """
        if self.expires_at is None:
            return None
        else:
            current_epoch_ts = time.time()
            if current_epoch_ts > (self.expires_at - 900):
                logger.debug("Access token has expired or will expire within the next 15 minutes. Refresh is needed.")
                return True
            else:
                logger.debug(f"Access token is still valid. Expires at: {self.expires_at}, Current time: {current_epoch_ts}")
                return False


class AuthProvider(ABC):
    @abstractmethod
    def get_credentials(self) -> Credentials | None:
        """
        Get credentials from their source and return them as a Credentials object.

        If no credentials are found, return None.

        Subclasses must implement this method to be used by the AuthResolver.
        :return: Credentials object or None if no credentials are found
        """


class AuthResolver:
    def __init__(self, providers: list[AuthProvider]):
        self.providers = providers

    def get_credentials(self) -> Credentials:
        logger.debug("Attempting to resolve credentials from providers.")
        for provider in self.providers:
            auth = provider.get_credentials()
            if auth:
                return auth
        return Credentials(credential_type=AuthenticationType.BROWSER)


class JwtEnvProvider(AuthProvider):
    def __init__(self, settings: SDKSettings):
        self._settings = settings

    def get_credentials(self):
        access_token = self._settings.access_token
        token_expiration = self._settings.token_expiration
        account_id = self._settings.account_id

        token_data = {}
        if access_token:
            token_data["access_token"] = access_token.get_secret_value()

        if token_expiration:
            token_data["expires_at"] = int(token_expiration)

        if account_id:
            token_data["account_id"] = str(account_id)

        if not token_data:
            return None

        token_data["credential_type"] = AuthenticationType.PASS_THROUGH
        return Credentials(**token_data)


class ClientCredentialsEnvProvider(AuthProvider):
    def __init__(self, settings: SDKSettings):
        self._settings = settings

    def get_credentials(self):
        client_id = self._settings.client_id
        client_secret = self._settings.client_secret
        account_id = self._settings.account_id

        if client_id and client_secret:
            return Credentials(
                client_id=client_id.get_secret_value(),
                client_secret=client_secret.get_secret_value(),
                account_id=account_id,
                credential_type=AuthenticationType.MACHINE_TO_MACHINE,
            )
        return None


class CredentialsFileProfileProvider(AuthProvider):
    """
    Reads credentials from a named profile section in ~/.komodo/credentials.
    Only supports service principal credentials (client_id/client_secret).
    Expects an INI file with sections like:
    [profile-name]
    client_id=...
    client_secret=...
    account_id=...
    account_slug=...
    """

    def __init__(self, settings: SDKSettings, profile_name: str):
        self._settings = settings
        self._profile_name = profile_name

    def get_credentials(self):
        credentials_path = os.path.expanduser("~/.komodo/credentials")
        if not os.path.exists(credentials_path):
            logging.warning(f"Credentials file not found at {credentials_path}. Please run `komodo login` to create it.")
            return None

        config = configparser.ConfigParser()
        try:
            config.read(credentials_path)
        except configparser.DuplicateSectionError as e:
            raise ValueError(f"Duplicate section found in credentials file at {credentials_path}. Please ensure each section header appears only once. Error: {e}") from e
        except Exception as e:
            logging.warning(f"Could not read credentials file: {e}")
            return None

        if self._profile_name not in config:
            logging.warning(f"Profile section [{self._profile_name}] not found in credentials file.")
            return None

        section = config[self._profile_name]
        client_id = section.get("client_id")
        client_secret = section.get("client_secret")
        account_id = section.get("account_id")
        account_slug = section.get("account_slug")

        if client_id and client_secret:
            creds_kwargs = {
                "client_id": client_id,
                "client_secret": client_secret,
                "credential_type": AuthenticationType.MACHINE_TO_MACHINE,
            }
            if account_id:
                creds_kwargs["account_id"] = account_id
            if account_slug:
                creds_kwargs["account_slug"] = account_slug
            return Credentials(**creds_kwargs)
        
        return None


class CredentialsFileDefaultProvider(AuthProvider):
    """
    Reads credentials from the [default] section in ~/.komodo/credentials.
    Supports JWT token credentials.
    Expects an INI file with sections like:
    [default]
    token=...
    token_expiration=...
    account_id=...
    account_slug=...
    """

    def __init__(self, settings: SDKSettings):
        self._settings = settings

    def get_credentials(self):
        credentials_path = os.path.expanduser("~/.komodo/credentials")
        if not os.path.exists(credentials_path):
            return None

        config = configparser.ConfigParser()
        try:
            config.read(credentials_path)
        except configparser.DuplicateSectionError as e:
            raise ValueError(f"Duplicate section found in credentials file at {credentials_path}. Please ensure each section header (e.g., [default]) appears only once. Error: {e}") from e
        except Exception as e:
            logging.warning(f"Could not read credentials file: {e}")
            return None

        section_name = "default"
        if section_name not in config:
            logging.warning(f"Default section [{section_name}] not found in credentials file.")
            return None

        section = config[section_name]
        token = section.get("token")
        expires_at_str = section.get("token_expiration")
        account_id = section.get("account_id")
        account_slug = section.get("account_slug")

        if token and expires_at_str:
            try:
                expires_at = int(expires_at_str)
            except ValueError:
                logging.warning(f"Invalid expiration timestamp in [{section_name}]: {expires_at_str}")
                return None
            
            creds_kwargs = {
                "access_token": token,
                "expires_at": expires_at,
            }
            if account_id:
                creds_kwargs["account_id"] = account_id
            if account_slug:
                creds_kwargs["account_slug"] = account_slug
            return Credentials(**creds_kwargs)
        
        return None

    def set_credentials(self, env: str, tokens: dict, file_path: str = None):
        """
        Write the provided tokens to the credentials file under the [default] section.
        Only update the fields provided in tokens, preserving any existing fields in the section.
        Also update/create a [session] section with komodo_environment set to the last written environment.
        """
        credentials_path = file_path or os.path.expanduser("~/.komodo/credentials")
        config = configparser.ConfigParser()
        if os.path.exists(credentials_path):
            try:
                config.read(credentials_path)
            except configparser.DuplicateSectionError as e:
                raise ValueError(f"Duplicate section found in credentials file at {credentials_path}. Please ensure each section header appears only once. Error: {e}") from e
        else:
            os.makedirs(os.path.dirname(credentials_path), exist_ok=True)

        section_name = "default"
        if section_name not in config:
            config.add_section(section_name)
        # Update with new tokens
        for key, value in tokens.items():
            if key == "access_token":
                config.set(section_name, "token", value)
            elif key == "expires_at":
                config.set(section_name, "token_expiration", str(value))
            elif key == "account_id":
                config.set(section_name, "account_id", str(value))
            elif key == "account_slug":
                config.set(section_name, "account_slug", str(value))

        # Add/update session section with environment
        if "session" not in config:
            config.add_section("session")
        config.set("session", "komodo_environment", env)

        with open(credentials_path, "w") as f:
            config.write(f)
        return credentials_path


class Session:
    def __init__(self, access_token=None, client_id=None, client_secret=None, environment: str | None = None, profile: str | None = None):
        """
        Initialize a Session object with authentication credentials.

        Credential resolution inspired by https://github.com/boto/botocore/blob/develop/botocore/credentials.py

        :param access_token: JWT token
        :param client_id: Client ID for machine-to-machine authentication
        :param client_secret: Client secret for machine-to-machine authentication
        :param environment: Environment name (integration or production)
        :param profile: Named profile to use from credentials file (e.g., "profile-1", "default")
        """

        self._logger = logging.getLogger(__name__)

        if (client_id is not None and client_secret is not None) and (access_token is not None):
            raise ValueError("Cannot pass in both access_token and client_id/client_secret")
        
        if profile is not None and (access_token is not None or client_id is not None or client_secret is not None):
            raise ValueError("Cannot specify profile along with explicit credentials (access_token, client_id, or client_secret)")

        self._access_token = access_token
        self._client_id = client_id
        self._client_secret = client_secret
        self._profile = profile
        self._refresh_lock = threading.Lock()

        if environment:
            self._settings = SDKSettings(environment=environment)
        else:
            self._settings = SDKSettings()

        self._token_data = self.load_credentials()
        
        # Update instance attributes from loaded credentials (for profile-based credentials)
        if self._token_data and not self._access_token and not (self._client_id and self._client_secret):
            if self._token_data.client_id:
                self._client_id = self._token_data.client_id
            if self._token_data.client_secret:
                self._client_secret = self._token_data.client_secret

    def load_credentials(self) -> Credentials:
        """
        Check each of the providers in order and return the first set of credentials found.

        If the access token or client credentials were provided during initialization, then use those and
        only retrieve the stored account ID if there is one.
        """
        if self._access_token or (self._client_id and self._client_secret):
            if self._access_token:
                credential_type = AuthenticationType.PASS_THROUGH
            else:
                credential_type = AuthenticationType.MACHINE_TO_MACHINE
            return Credentials(
                access_token=self._access_token,
                client_id=self._client_id,
                client_secret=self._client_secret,
                credential_type=credential_type,
            )

        providers = [
            JwtEnvProvider(self._settings),
            ClientCredentialsEnvProvider(self._settings),
        ]
        
        # If profile is specified, use profile provider first; otherwise use default profile provider
        # If profile is 'default', use CredentialsFileDefaultProvider instead of CredentialsFileProfileProvider
        if self._profile is not None:
            if self._profile == "default":
                providers.insert(0, CredentialsFileDefaultProvider(self._settings))
            else:
                providers.insert(0, CredentialsFileProfileProvider(self._settings, self._profile))
        else:
            providers.append(CredentialsFileDefaultProvider(self._settings))
        
        resolver = AuthResolver(providers)
        creds = resolver.get_credentials()
        return creds

    @property
    def access_token(self) -> str | None:
        """
        Get the access token for the AuthSession. Represents a JWT used as the Authorization Bearer header in SDK HTTP
            requests.

        :return: JWT token
        """

        if self._token_data.refresh_needed():
            # Put a threading lock on the _token_data object to prevent multiple threads from trying to refresh the token at once
            # If the lock can't be acquired, block until it is released and the acquire it
            if self._refresh_lock.acquire(False):
                try:
                    if not self._token_data.refresh_needed():
                        return self._token_data.access_token
                    self.connect()
                    return self._token_data.access_token
                finally:
                    self._refresh_lock.release()
            else:
                with self._refresh_lock:
                    if not self._token_data.refresh_needed():
                        return self._token_data.access_token
                    self.connect()
        return self._token_data.access_token

    @property
    def expires_at(self) -> int | None:
        """
        The epoch timestamp when the access token expires

        :return: Epoch timestamp when the access token expires or None
        """
        return self._token_data.expires_at

    @property
    def account_id(self) -> str | None:
        return self._token_data.account_id if self._token_data else None

    def connect(self):
        # The integration /token endpoint is only available behind the rbac proxy url, aka /platform-rbac-stage/
        rbac_url = self._settings.komodo_external if self._settings.environment == "production" else f"{self._settings.komodo_external}/platform-rbac-stage"
        oauth = OAuthFlows(domain=self._settings.komodo_idp_domain, rbac_url=rbac_url, logger=self._logger)

        tokens = None
        match self._token_data.credential_type:
            case AuthenticationType.BROWSER:
                if self._token_data.refresh_needed():
                    self._logger.warning("Access token will expire soon. Please re-authenticate.")
                else:
                    tokens = oauth.device_authorization(
                        self._settings.komodo_client_id,
                        self._settings.komodo_idp_audience,
                        KOMODO_IDP_API_SCOPE,
                    )
            case AuthenticationType.MACHINE_TO_MACHINE:
                if self._token_data.access_token is not None:
                    if self._token_data.refresh_needed():
                        tokens = oauth.client_credentials(
                            self._client_id,
                            self._client_secret,
                        )
                else:
                    tokens = oauth.client_credentials(
                        self._client_id,
                        self._client_secret,
                    )

        if tokens is not None:
            # Preserve account_id and account_slug from original credentials when updating tokens
            creds_kwargs = {**tokens, "credential_type": self._token_data.credential_type}
            if self._token_data.account_id:
                creds_kwargs["account_id"] = self._token_data.account_id
            if self._token_data.account_slug:
                creds_kwargs["account_slug"] = self._token_data.account_slug
            self._token_data = Credentials(**creds_kwargs)



import os
from enum import Enum
from typing import Dict


class Vendors(str, Enum):
    SNOWFLAKE = "snowflake"
    SNOWFLAKE_NATIVE = "snowflake-native"
    SNOWFLAKE_ROTATE_KEYS = "snowflake-rotate-keys"

    @staticmethod
    def values() -> Dict[str, Enum]:
        return Vendors._value2member_map_


class Environment(Enum):
    PROD = "prod"
    DEV = "dev"
    STAGE = "stage"
    LOCAL = "local"


# TODO: Update the host config to the proper DNS for each environment
KOMODO_PROXY_HOST = {
    Environment.PROD: "connector-gateway.onkomodo.com",
    Environment.DEV: "connector-gateway.staging.onkomodo.com",
    Environment.STAGE: "connector-gateway.staging.onkomodo.com",
    Environment.LOCAL: "127.0.0.1",
}

# This port is only used for local development.
# For all remote requests, we remove the port before sending the request and only rely on the dns
DEV_KOMODO_PROXY_SERVICE_DEFAULT_PORT = os.getenv("DEV_KOMODO_PROXY_SERVICE_DEFAULT_PORT", "8000")

VENDOR_CONNECTION_STRING_MAP: Dict[Vendors, str] = {
    Vendors.SNOWFLAKE: "snowflake",
    Vendors.SNOWFLAKE_NATIVE: "snowflake",
    Vendors.SNOWFLAKE_ROTATE_KEYS: "snowflake",
}
# TODO: Remove komodohealth-komodohealthent_compute_dev after testing
SNOWFLAKE_MAIN_ACCOUNTS = ("komodohealthent", "komodohealth")



from typing import Any, Dict, Optional

from sqlalchemy.engine import CreateEnginePlugin
from sqlalchemy.engine.url import URL

from komodo._vendored.komodo_sqlalchemy.dialect import KomodoDriverDialect


class KomodoEnginePlugin(CreateEnginePlugin):
    def __init__(self, url: URL, *kwargs: Optional[Dict[str, Any]]) -> None:
        KomodoDriverDialect.connection_string = url

    def update_url(self, url: URL) -> URL:
        # Store the connection string in KomodoDriverDialect
        KomodoDriverDialect.connection_string = url
        # Return the URL without modification
        return url



## This is an auto-generated file. Do not edit this file directly.
import asyncio
import importlib.metadata
import logging
import os
import platform
import sys

from types import TracebackType
from typing import Optional, Type

import httpx

from komodo.auth.Session import Session

from komodo.api_client import ApiClient
from komodo.configuration import Configuration

from komodo.api.iam_api import IamApi


class Client:
    def __init__(
        self,
        account_id: str = None,
        auth_session: Session | None = None,
        **kwargs,
    ):
        """

        :param account_id
        :param auth_session: optional Komodo Platform session
        """
        self._logger = logging.getLogger(__name__)
        self._auth_session = auth_session if auth_session is not None else Session()

        self._auth_session.connect()

        self.x_account_id = account_id if account_id is not None else self._auth_session.account_id

        self._client_session = self.komodo_api_client()

    def komodo_api_client(self) -> ApiClient:
        access_token = self._auth_session._token_data.access_token
        configuration = Configuration(
            api_key={"JWT": access_token, "HTTPBearer": access_token, "x-account-id": self.x_account_id}, api_key_prefix={"JWT": "Bearer", "HTTPBearer": "Bearer"}, access_token=access_token
        )
        configuration.refresh_api_key_hook = self.refresh_api_key_hook

        client = ApiClient(configuration, header_name="X-API-Version", header_value="2024-10-01")

        if "HOSTNAME" in os.environ:
            client.set_default_header("X-KH-Compute-ContainerID", os.environ["HOSTNAME"])

        version = importlib.metadata.version("komodo")
        system_info = self.get_system_info()

        user_agent = f"KomodoSDK/{version} ({system_info})"
        client.user_agent = user_agent
        return client

    @property
    def logger(self):
        """
        SDK session logger
        """
        return self._logger

    @property
    def auth(self) -> Session:
        """
        SDK auth session
        """
        return self._auth_session

    @property
    def session(self) -> httpx.AsyncClient:
        """
        SDK HTTP session
        """
        return self._client_session

    @classmethod
    async def create(
        cls,
        account_id: str = None,
        auth_session: Session | None = None,
        **kwargs,
    ) -> "Client":
        """
        Create a Client instance from optional identifier and Komodo Platform auth
        session.

        The HTTP session is automatically set.

        :param account_id:
        :param auth_session:
        :param kwargs:
        :return:
        """
        self = Client(account_id, auth_session, **kwargs)
        return await self.__aenter__()

    async def close(self):
        await self.__aexit__(None, None, None)

    async def __aenter__(self) -> "Client":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self._logger.debug(f"Closing Client")
        await asyncio.sleep(0.250)  # https://docs.aiohttp.org/en/stable/client_advanced.html#graceful-shutdown
        await self._client_session.rest_client_async.close()

    @property
    def iam(self) -> IamApi:
        return IamApi(api_client=self._client_session)

    @staticmethod
    def get_system_info() -> str:
        """
        Get system information including Python version and system architecture.

        Returns:
            str: Formatted string with system information
        """
        python_version = sys.version.split()[0]
        system_name = platform.system()
        machine_arch = platform.machine()

        return f"Python {python_version}; {system_name} {machine_arch}"

    def refresh_api_key_hook(self, configuration):
        """
        Update the configuration's access token by accessing the auth session's
        access_token property and setting the configuration's api_key. The property
        automatically refreshes the token if it is expired.

        Args:
            configuration: The configuration object to update.
        """

        access_token = self._auth_session.access_token
        configuration.access_token = access_token
        configuration.api_key = {"JWT": access_token, "HTTPBearer": access_token, "x-account-id": self.x_account_id}



# type: ignore
from typing import Any, Dict, List, Optional, Tuple

from komodo._vendored.komodo_connector.connection import Connection
from komodo._vendored.komodo_connector.connector import connect as komodo_connect
from snowflake.sqlalchemy import dialect as SnowflakeDialect
from sqlalchemy.engine import default, reflection
from sqlalchemy.engine.url import URL

VENDOR_DIALECTS = {
    "snowflake": SnowflakeDialect,
}


class KomodoDialectMeta(type):
    def __getattr__(cls, name: str) -> Any:
        vendor_dialect = cls.get_dialect_class_from_url(cls.connection_string)
        if vendor_dialect is None:
            raise AttributeError(f"{cls.__name__} object has no attribute {name}")

        return getattr(vendor_dialect, name)


class KomodoDriverDialect(default.DefaultDialect, metaclass=KomodoDialectMeta):
    # A Dialect class needs to be configured at build time for a SQLAlchemy plugin.
    #
    # However, since we need to dynamically determine the vendor to inherit from at runtime,
    # we use a combination of a sqlAlchemy engine plugin that populates the class attribute
    # 'connection_string' at runtime, and a metaclass that uses this class attribute to
    # dynamically sets a __getattr__ from dialect of the vendor set in the connection string.
    #
    # We cannot dynamically set properties at runtime in the class methods because sqlAlchemy
    # looks up class attributes that need to be fetched from the vendor Dialect before any
    # KomodoDriverDialect instances are created, so none of the class method bodies are *executed*
    # until after the class is instantiated.
    #
    # This allows us to use the same Dialect class for all vendors, and dynamically inherit
    # from the correct vendor dialect at runtime.

    connection_string: URL = None  # Populated at runtime, do not set here
    optional_connection_params: Tuple[str, ...] = ("protocol", "env", "komodo_host", "app_id", "role", "warehouse")
    dialect_connection_args_key = "dialect_connection_args"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.connection_string is None:
            raise ValueError("Connection string cannot be None")

        vendor_dialect_class = self.get_dialect_class_from_url(self.connection_string)
        self._delegate = vendor_dialect_class(*args, **kwargs)

    def __getattr__(self, name: str) -> Any:
        if name.startswith("_"):
            raise AttributeError(f"{self.__class__.__name__} object has no attribute {name}")

        if callable(attr := getattr(self._delegate, name)):

            def wrapper(*args, **kwargs):
                return attr(*args, **kwargs)

            return wrapper
        else:
            return attr

    @classmethod
    def get_dialect_class_from_url(cls, engine_url: Optional[URL]) -> Optional[default.DefaultDialect]:
        if engine_url is None:
            raise ValueError("Engine URL cannot be None")

        if vendor := (engine_url.query.get("dialect") or engine_url.username.split("+", 1)[0]):
            try:
                return VENDOR_DIALECTS[vendor]
            except KeyError as e:
                raise NotImplementedError(f"Vendor {vendor} is not supported") from e
        raise NotImplementedError("Vendor not specified")

    def create_connect_args(self, url: URL) -> Tuple[List[Any], Dict[str, str]]:
        return self._delegate.create_connect_args(url)

    def connect(self, *cargs: Dict[str, str], **cparams: Dict[str, str]) -> Connection:
        connect_params = {param: cparams[param] for param in self.optional_connection_params if param in cparams}
        # Need to account for if the params are in the query string
        query_string_params = {
            param: self.connection_string.query.get(param)
            for param in self.optional_connection_params
            if self.connection_string.query.get(param)
        }

        if query_arg := self.connection_string.query.get("dialect"):
            # SQLAlchemy doesn't support multiple dialects in the connection string, so we
            # need to add the dialect to the username and parse it out later in the Komodo driver
            username = f"{query_arg}+{self.connection_string.username}"
        else:
            username = self.connection_string.username

        self.connection_string = URL(
            drivername=self.connection_string.drivername,
            username=username,
            password=self.connection_string.password,
            host=self.connection_string.host,
            port=self.connection_string.port,
            database=self.connection_string.database,
            query=connect_params | query_string_params,
        )
        connection = komodo_connect(
            self.connection_string.render_as_string(hide_password=False),
            **cparams.get(self.dialect_connection_args_key, {}),
        )
        return connection

    @reflection.cache
    def _current_database_schema(self, connection, **kw):
        return self._delegate._current_database_schema(connection, **kw)

    @reflection.cache
    def get_indexes(self, connection, table_name, schema=None, **kw):
        return self._delegate.get_indexes(connection, table_name, schema, **kw)

    @reflection.cache
    def get_check_constraints(self, connection, table_name, schema, **kw):
        return self._delegate.get_check_constraints(connection, table_name, schema, **kw)

    @reflection.cache
    def _get_schema_primary_keys(self, connection, schema, **kw):
        return self._delegate._get_schema_primary_keys(connection, schema, **kw)

    @reflection.cache
    def _get_schema_unique_constraints(self, connection, schema, **kw):
        return self._delegate._get_schema_unique_constraints(connection, schema, **kw)

    @reflection.cache
    def _get_schema_foreign_keys(self, connection, schema, **kw):
        return self._delegate._get_schema_foreign_keys(connection, schema, **kw)

    @reflection.cache
    def _get_schema_columns(self, connection, schema, **kw):
        return self._delegate._get_schema_columns(connection, schema, **kw)

    @reflection.cache
    def _get_table_columns(self, connection, table_name, schema=None, **kw):
        return self._delegate._get_table_columns(connection, table_name, schema, **kw)

    @reflection.cache
    def get_table_names(self, connection, schema=None, **kw):
        return self._delegate.get_table_names(connection, schema, **kw)

    @reflection.cache
    def get_view_names(self, connection, schema=None, **kw):
        return self._delegate.get_view_names(connection, schema, **kw)

    @reflection.cache
    def get_view_definition(self, connection, view_name, schema=None, **kw):
        return self._delegate.get_view_definition(connection, view_name, schema, **kw)

    @reflection.cache
    def get_sequence_names(self, connection, schema=None, **kw):
        return self._delegate.get_sequence_names(connection, schema, **kw)

    @reflection.cache
    def get_schema_names(self, connection, **kw):
        return self._delegate.get_schema_names(connection, **kw)

    @reflection.cache
    def get_columns(self, connection, table_name, schema=None, **kw):
        return self._delegate.get_columns(connection, table_name, schema, **kw)

    @reflection.cache
    def get_pk_constraint(self, connection, table_name, schema=None, **kw):
        return self._delegate.get_pk_constraint(connection, table_name, schema, **kw)

    @reflection.cache
    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        return self._delegate.get_foreign_keys(connection, table_name, schema, **kw)

    @reflection.cache
    def has_table(self, connection, table_name, schema=None):
        return self._delegate.has_table(connection, table_name, schema)

    @reflection.cache
    def has_sequence(self, connection, sequence_name, schema=None):
        return self._delegate.has_sequence(connection, sequence_name, schema)
